'''
* Copyright 2015-2018 European Atomic Energy Community (EURATOM)
*
* Licensed under the EUPL, Version 1.1 or - as soon they
  will be approved by the European Commission - subsequent
  versions of the EUPL (the "Licence");
* You may not use this work except in compliance with the
  Licence.
* You may obtain a copy of the Licence at:
*
* https://joinup.ec.europa.eu/software/page/eupl
*
* Unless required by applicable law or agreed to in
  writing, software distributed under the Licence is
  distributed on an "AS IS" basis,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
  express or implied.
* See the Licence for the specific language governing
  permissions and limitations under the Licence.
'''


"""
vtk Interactor style classes for Calcam.

There's a lot more code in here than there needs to be / should be,
a lot of what's done in here should probably be moved to the GUI module.
However, because a lot of this code predates the Qt GUI, it was never moved out of here,
which is why the GUI window object and vtkinteractor object interact with
each other in a sadly messy way.

Written by Scott Silburn 
"""

import vtk
import numpy as np
from ..render import get_image_actor
from ..misc import rotate_3d

import time


class CalcamInteractorStyle3D(vtk.vtkInteractorStyleTerrain):
 
    def __init__(self,parent=None,viewport_callback=None,resize_callback=None,newpick_callback=None,cursor_move_callback=None,focus_changed_callback=None,refresh_callback=None,pre_move_callback=None):
        
        # Set callbacks for all the mouse controls

        self.AddObserver("LeftButtonPressEvent",self.on_left_click)
        self.AddObserver("RightButtonPressEvent",self.right_press)
        self.AddObserver("RightButtonReleaseEvent",self.right_release)
        self.AddObserver("MiddleButtonPressEvent",self.middle_press)
        self.AddObserver("MiddleButtonReleaseEvent",self.middle_release)
        self.AddObserver("MouseWheelForwardEvent",self.zoom_in)
        self.AddObserver("MouseWheelBackwardEvent",self.zoom_out)
        self.AddObserver("MouseMoveEvent",self.on_mouse_move)
        self.viewport_callback = viewport_callback
        self.pick_callback = newpick_callback
        self.resize_callback = resize_callback
        self.newpick_callback = newpick_callback
        self.cursor_move_callback = cursor_move_callback
        self.cursor_changed_callback = focus_changed_callback
        self.refresh_callback = refresh_callback
        self.focus_changed_callback = focus_changed_callback
        self.pre_move_callback = pre_move_callback
        self.image_actor = None
        self.image_resizer = None
        self.force_aspect = None
        self.im_aspect = None
        self.zoom_enabled=True
        self.projection = 'perspective'
        self.rightdrag_rotate = False
        self.control_sensitivity = 0.75
        self.cam_roll = 0.
        self.rmb_down = False
        self.mouse_delta = np.array([0,0])
        self.allow_focus_change = True



    # Do various initial setup things, most of which can't be done at the time of __init__
    def init(self):

        # Get the interactor object
        self.interactor = self.GetInteractor()

        # Some other objects from higher up which I need access to
        self.vtkwindow = self.interactor.GetRenderWindow()
        renderers = self.vtkwindow.GetRenderers()
        renderers.InitTraversal()
        self.renderer = renderers.GetNextItemAsObject()
        self.camera = self.renderer.GetActiveCamera()

        self.camera.SetViewUp(0,0,1)

        self.SetAutoAdjustCameraClippingRange(False)

        # Turn off any VTK responses to keyboard input (all necessary keyboard shortcuts etc are done in Q)
        self.interactor.RemoveObservers('KeyPressEvent')
        self.interactor.RemoveObservers('CharEvent')

        # Add observer for catching window resizing
        self.vtkwindow.AddObserver("ModifiedEvent",self.on_resize)

        self.extra_actors = []

        # Create a picker
        self.picker = vtk.vtkCellPicker()
        self.interactor.SetPicker(self.picker)

        # Variables
        self.cursors = {}
        self.next_cursor_id = 0
        self.focus_cursor = None
        self.legend = None
        self.xsection_coords = None
    

        # We will use this for converting from 3D to screen coords.
        self.vtk_coord_transformer = vtk.vtkCoordinate()
        self.vtk_coord_transformer.SetCoordinateSystemToWorld()


    def set_rmb_rotate(self,rmb_rotate):

        self.rightdrag_rotate = rmb_rotate


    def set_fov(self,fov):

        if self.projection == 'perspective':
            self.camera.SetViewAngle(fov)
        else:
            self.camera.SetParallelScale(fov/2.)


    def set_control_sensitivity(self,sensitivity):

        if sensitivity < 0 or sensitivity > 1.:
            raise ValueError('Sensitivity value must be between 0 and 1. ')
        else:
            self.control_sensitivity = sensitivity + 0.05


    def set_projection(self,projection,fov=None):

        if projection not in ['perspective','orthographic']:
            raise ValueError('Projection should be "perspective" or "orthographic"')

        if projection == self.projection:
            return

        if projection == 'orthographic':

            if fov is None:
                fov = np.abs(np.tan( 3.14159*self.camera.GetViewAngle()/360 ) * np.dot(self.camera.GetPosition(),self.camera.GetDirectionOfProjection()))

            self.camera.SetParallelProjection(True)
            self.camera.SetParallelScale(fov)


        elif projection == 'perspective':

            if fov is None:
                fov = np.abs(360*np.arctan(self.camera.GetParallelScale()/np.dot(self.camera.GetPosition(),self.camera.GetDirectionOfProjection()))/3.14159)

            self.camera.SetParallelProjection(False)
            self.camera.SetViewAngle(fov)

        self.projection = projection


    def set_upvec(self,upvec):

        self.camera_3d.SetViewUp(upvec)
        cam_roll = self.camera_3d.GetRoll()



    def get_projection(self):
        return self.projection

    # Middle click + drag to pan
    def middle_press(self,obj,event):
        self.orig_dist = self.camera.GetDistance()
        self.camera.SetDistance(self.control_sensitivity)
        self.OnMiddleButtonDown()


    def middle_release(self,obj,event):
        self.OnMiddleButtonUp()
        self.camera.SetDistance(self.orig_dist)
        self.on_cam_moved()


    # On the CAD view, right click+drag to rotate (usually on left button in this interactorstyle)
    def right_press(self,obj,event):
        self.orig_dist = self.camera.GetDistance()
        if self.rightdrag_rotate:
            self.camera.SetDistance( np.sqrt( np.sum( np.array(self.camera.GetPosition())**2)) )
        else:
            self.camera.SetDistance(0.01)
        self.rmb_down = True
        self.OnLeftButtonDown()


    def right_release(self,obj,event):
        self.rmb_down = False
        self.OnLeftButtonUp()
        self.camera.SetDistance(self.orig_dist)
        self.on_cam_moved()



    def zoom_in(self,obj,event):

        # If ctrl + scroll, change the camera FOV
        if self.interactor.GetControlKey():
            if self.zoom_enabled:
                if self.projection == 'perspective':
                    self.camera.SetViewAngle(max(self.camera.GetViewAngle()*(1 - self.control_sensitivity*0.25),1))
                elif self.projection == 'orthographic':
                    self.camera.SetParallelScale(max(0.01,self.camera.GetParallelScale()*(1 - self.control_sensitivity*0.25)))

        # Otherwise, move the camera forward.
        else:
            orig_dist = self.camera.GetDistance()
            self.camera.SetDistance(1.3 * self.control_sensitivity)
            self.camera.Dolly(1.1)
            self.camera.SetDistance(orig_dist)

        # Update cursor sizes depending on their distance from the camera,
        # so they're all comfortably visible and clickable.
        self.on_cam_moved()
   


    def zoom_out(self,obj,event):

        # If ctrl + scroll, change the camera FOV
        if self.interactor.GetControlKey():
            if self.zoom_enabled:
                if self.projection == 'perspective':
                    self.camera.SetViewAngle(min(self.camera.GetViewAngle()*(1 + self.control_sensitivity*0.25),110.))
                elif self.projection == 'orthographic':
                    self.camera.SetParallelScale(min(500,self.camera.GetParallelScale()*(1 + self.control_sensitivity*0.25)))

        # Otherwise, move the camera backward.
        else:
            orig_dist = self.camera.GetDistance()
            self.camera.SetDistance(1.3 * self.control_sensitivity)
            self.camera.Dolly(0.9)
            self.camera.SetDistance(orig_dist)

        # Update cursor sizes so they're all well visible:
        self.on_cam_moved()


    def on_cam_moved(self):
        self.update_cursor_style(refresh=False)
        self.update_clipping()

        if self.viewport_callback is not None:
            self.viewport_callback()


    def set_xsection(self,xsection_coords):

        if xsection_coords is not None:
            self.xsection_coords = xsection_coords
        else:
            self.xsection_coords = None

        self.update_clipping()
    
    def get_xsection(self):
        return self.xsection_coords



    def get_cursor_coords(self,cursor_id):

        return self.cursors[cursor_id]['cursor3d'].GetFocalPoint()


    def add_extra_actor(self,actor):

        if actor not in self.extra_actors:
            self.extra_actors.append(actor)
            self.renderer.AddActor(actor)


    def remove_extra_actor(self,actor):

        if actor in self.extra_actors:
            self.extra_actors.remove(actor)
            self.renderer.RemoveActor(actor)


    def update_clipping(self):

        self.renderer.ResetCameraClippingRange()

        if self.xsection_coords is not None:
            normal_range = self.camera.GetClippingRange()
            cam_to_xsec = self.xsection_coords - np.array(self.camera.GetPosition())
            cam_view_dir = self.camera.GetDirectionOfProjection()
            dist = max(normal_range[0],np.dot(cam_to_xsec,cam_view_dir))
            self.camera.SetClippingRange(dist,normal_range[1])

        if self.refresh_callback is not None:
            self.refresh_callback()


    # Left click to move a point or add a new point
    def on_left_click(self,obj,event):

        ctrl_pressed = self.interactor.GetControlKey()

        # These will be the variables we return. If the user clicked in free space they will stay None.
        clicked_cursor = None
        pickcoords = None

        # Do a pick with our picker object
        clickcoords = self.interactor.GetEventPosition()
        
        for actor in self.extra_actors:
            self.renderer.RemoveActor(actor)
        
        retval = self.picker.Pick(clickcoords[0],clickcoords[1],0,self.renderer)

        # If something was successfully picked, find out what it was...
        if retval != 0:

            pickedpoints = self.picker.GetPickedPositions()

            # If more than 1 point is within the picker's tolerance,
            # use the one closest to the camera (this is most intuitive)
            npoints = pickedpoints.GetNumberOfPoints()
            dist_fromcam = []
            campos = self.camera.GetPosition()

            for i in range(npoints):
                point = pickedpoints.GetPoint(i)
                dist_fromcam.append(np.sqrt( (campos[0] - point[0])**2 + (campos[1] - point[1])**2 + (campos[2] - point[2])**2 ))

            _, idx = min((val, idx) for (idx, val) in enumerate(dist_fromcam))

            pickcoords = pickedpoints.GetPoint(idx)

            # If the picked point is within 1.5x the cursor radius of any existing point,
            # say that the user clicked on that point
            dist = 7
            for cid,cursor in self.cursors.items():

                if cid == self.focus_cursor:
                    continue

                # Get the on-screen position of this cursor
                self.vtk_coord_transformer.SetValue(cursor['cursor3d'].GetFocalPoint())
                cursorpos = self.vtk_coord_transformer.GetComputedDisplayValue(self.renderer)
                

                dist_from_cursor = np.sqrt( (cursorpos[0] - clickcoords[0])**2 + (cursorpos[1] - clickcoords[1])**2 )
                if dist_from_cursor < dist:
                        clicked_cursor = cid
                        dist = dist_from_cursor



            # If they held CTRL, we send a new pick callback
            if (ctrl_pressed or self.focus_cursor is None) and clicked_cursor is None:

                if self.newpick_callback is not None:
                    self.newpick_callback(pickcoords)

            else:

                # Otherwise, if they clicked an existing cursor, change the focus to it
                if clicked_cursor is not None and self.allow_focus_change:

                    self.set_cursor_focus(clicked_cursor)

                    if self.focus_changed_callback is not None:
                        self.focus_changed_callback(clicked_cursor)


                # of if they didn't click another cursor, move the current cursor
                # to where they clicked
                elif self.focus_cursor is not None:
                    if self.pre_move_callback is not None:
                        self.pre_move_callback()
                    self.set_cursor_coords(self.focus_cursor,pickcoords)


        for actor in self.extra_actors:
            self.renderer.AddActor(actor)

            if self.refresh_callback is not None:
                self.refresh_callback()



    def set_cursor_coords(self,cursor_id,coords):

        self.cursors[cursor_id]['cursor3d'].SetFocalPoint(coords)
        self.update_cursor_style()

        if self.cursor_move_callback is not None:
            self.cursor_move_callback(self.focus_cursor,coords)


    def update_cursor_style(self,refresh=True):

        campos = self.camera.GetPosition()
        for cid,cursor in self.cursors.items():

            position = cursor['cursor3d'].GetFocalPoint()

            dist_to_cam = np.sqrt( (campos[0] - position[0])**2 + (campos[1] - position[1])**2 + (campos[2] - position[2])**2 )

            if cid == self.focus_cursor:
                colour = (0,0.8,0)
                linewidth = 3
                size = 0.025
            else:
                if cursor['colour'] is not None:
                    colour = cursor['colour']
                else:
                    colour = (0.8,0,0)
                linewidth = 2
                size = 0.0125

            # Cursor size scales with camera FOV to maintain size on screen.
            size = size * (self.camera.GetViewAngle()/75)

            cursor['cursor3d'].SetModelBounds([position[0]-size*dist_to_cam,position[0]+size*dist_to_cam,position[1]-size*dist_to_cam,position[1]+size*dist_to_cam,position[2]-size*dist_to_cam,position[2]+size*dist_to_cam])
            cursor['actor'].GetProperty().SetColor(colour)
            cursor['actor'].GetProperty().SetLineWidth(linewidth)

        if self.refresh_callback is not None and refresh:
            self.refresh_callback()



    # Defocus cursors for a given point pair
    def set_cursor_focus(self,cursor_id):

        if cursor_id is not None:
            if cursor_id not in self.cursors.keys():
                raise ValueError('No cursor with ID {:d}'.format(cursor_id))

        self.focus_cursor = cursor_id
        self.update_cursor_style()


    def get_cursor_focus(self):
        return self.focus_cursor


    def add_cursor(self,coords):

        # Create new cursor, mapper and actor
        new_cursor_id = self.next_cursor_id
        self.cursors[new_cursor_id] = {'cursor3d':vtk.vtkCursor3D(),'actor':vtk.vtkActor(),'colour':None}
        self.next_cursor_id += 1

        # Some setup of the cursor
        self.cursors[new_cursor_id]['cursor3d'].XShadowsOff()
        self.cursors[new_cursor_id]['cursor3d'].YShadowsOff()
        self.cursors[new_cursor_id]['cursor3d'].ZShadowsOff()
        self.cursors[new_cursor_id]['cursor3d'].OutlineOff()
        self.cursors[new_cursor_id]['cursor3d'].SetTranslationMode(1)
        self.cursors[new_cursor_id]['cursor3d'].SetFocalPoint(coords[0],coords[1],coords[2])


        # Mapper setup
        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputConnection(self.cursors[new_cursor_id]['cursor3d'].GetOutputPort())
    
        # Actor setup
        self.cursors[new_cursor_id]['actor'].SetMapper(mapper)

        # Add new cursor to screen
        self.renderer.AddActor(self.cursors[new_cursor_id]['actor'])

        self.update_cursor_style()

        return new_cursor_id
  

    def remove_cursor(self,cursor_id):

        try:
            cursor = self.cursors.pop(cursor_id)
            self.renderer.RemoveActor(cursor['actor'])
            if cursor_id == self.focus_cursor:
                self.focus_cursor = None
                self.update_cursor_style()
        except KeyError:
            raise ValueError('No cursor with ID {:d}'.format(cursor_id))


    def on_resize(self,obj=None,event=None):

        vtk_size = self.vtkwindow.GetSize()

        vtk_aspect = float(vtk_size[1]) / float(vtk_size[0])

        if self.force_aspect is not None:
            # Camera view wider than VTK panel
            if self.force_aspect <= vtk_aspect:
                h = self.force_aspect / vtk_aspect
                self.renderer.SetViewport([0.,0.5-h/2.,1.0,0.5+h/2.])

            # Camera view taller than VTK panel
            elif self.force_aspect > vtk_aspect:
                w = vtk_aspect / self.force_aspect
                self.renderer.SetViewport([0.5-w/2.,0.,0.5+w/2.,1.])

        # Sizing of the legend
        if self.legend is not None:

            legend_offset_y = 0.02
            legend_scale = 0.03

            legend_offset_x = legend_offset_y*vtk_size[1] / vtk_size[0]
            legend_pad_y = 30./vtk_size[1]
            legend_pad_x = 20./vtk_size[0]

            legend_height = legend_pad_y + legend_scale * self.n_legend_items
            abs_height = legend_scale * vtk_size[1]
            width_per_char = abs_height * 0.5
            legend_width = legend_pad_x + (width_per_char * self.longest_name)/vtk_size[0]

            self.legend.GetPosition2Coordinate().SetCoordinateSystemToNormalizedDisplay()
            self.legend.GetPositionCoordinate().SetCoordinateSystemToNormalizedDisplay()

            # Set Legend Size
            self.legend.GetPosition2Coordinate().SetValue(legend_width,legend_height )
            # Set Legend position
            self.legend.GetPositionCoordinate().SetValue(1 - legend_offset_x - legend_width, legend_offset_y)


        # Sizing of the overlay image
        if self.image_actor is not None:
            viewport = self.renderer.GetViewport()
            h = int(vtk_size[1] * (viewport[3] - viewport[1]))
            w = int(h / self.im_aspect)
            self.image_resizer.SetOutputDimensions(w,h,1)

        if self.resize_callback is not None:
            self.resize_callback(vtk_size)


    def set_legend(self,legend_items):

        if self.legend is not None:
            self.renderer.RemoveActor(self.legend)

        if len(legend_items) > 0:

            self.longest_name = 0
            for entry in legend_items:
                if len(entry[0]) > self.longest_name:
                    self.longest_name = len(entry[0])

            self.n_legend_items = len(legend_items)

            legend = vtk.vtkLegendBoxActor()
            legend.SetNumberOfEntries(len(legend_items))

            for i,entry in enumerate(legend_items):
                legend.SetEntryString(i,entry[0])
                legend.SetEntryColor(i,entry[1])

 
            legend.UseBackgroundOn()
            legend.SetBackgroundColor((0.1,0.1,0.1))
            legend.SetPadding(9)
            self.legend = legend

            self.renderer.AddActor(self.legend)

            self.on_resize()


    def on_mouse_move(self,obj=None,event=None):

        rerender = False

        # Apply the mouse sensitivity by adjusting the apparent amount the mouse has moved
        # before calling the VTK methods
        # -----------------------------------
        xy = np.array(self.interactor.GetEventPosition())
        lastxy = np.array(self.interactor.GetLastEventPosition())

        delta = ( lastxy - xy ) * self.control_sensitivity + self.mouse_delta

        self.mouse_delta = np.mod(np.abs(delta),1) * np.sign(delta)
        delta = np.trunc(delta).astype(int)
        self.interactor.SetLastEventPosition(int(xy[0]+delta[0]),int(xy[1]+delta[1]))
        # -----------------------------------

        # Correctly handle mouse interaction:
        view_direction = self.camera.GetDirectionOfProjection()

        # If doing anything other than Right Click + Ctrl + Drag, we can just use 
        # the existing VTK method
        if not (self.rmb_down and self.interactor.GetControlKey()):

            self.OnMouseMove()

            # Make sure we maintain the set camera roll.
            if self.rmb_down and np.abs(self.cam_roll) > 0:
                self.set_roll(self.cam_roll)

        # Right click + drag rolls the camera.
        # This first case for if we're not looking vertically
        elif np.abs(view_direction[2]) < 0.97:
            lastxy = xy + delta
            cc = np.array(self.vtkwindow.GetSize())/2.
            delta_theta = (np.arctan( (xy[0] - cc[0])/(xy[1] - cc[1]) ) - np.arctan( (lastxy[0] - cc[0])/(lastxy[1] - cc[1]) ) )

            if np.abs(delta_theta) > 3:
                delta_theta = delta_theta - 3.14159*np.sign(delta_theta)

            roll = self.cam_roll - 180*delta_theta/3.14159
            if roll > 180:
                roll = roll - 360
            elif roll < -180:
                roll = roll + 360

            self.set_roll(roll)

        # If we are looking vertically, pretend the mouse has only moved horizontally
        # and call the regular handler. Believe it or not this enables consistent and smooth-ish rotation.
        else:
            self.interactor.SetLastEventPosition(int(xy[0]+delta[0]),xy[1])
            self.OnMouseMove()

        if self.xsection_coords is not None:
            self.update_clipping()


    def set_roll(self,roll):

        self.cam_roll = roll
        view_direction = self.camera.GetDirectionOfProjection()
        if np.abs(view_direction[2]) < 0.999:
            z_projection = np.array([ -view_direction[0]*view_direction[2], -view_direction[1]*view_direction[2], 1-view_direction[2]**2 ])
            upvec = rotate_3d(z_projection,view_direction,self.cam_roll)
            self.camera.SetViewUp(upvec)
            roll = self.camera.GetRoll()
            tar = self.camera.GetFocalPoint()
            self.camera.SetViewUp(0,0,1)
            self.camera.SetRoll(roll)
            self.camera.SetFocalPoint(tar)

            if self.refresh_callback is not None:
                self.refresh_callback()


    def set_overlay_image(self,im_array):

        if self.image_actor is not None:
            self.renderer.RemoveActor(self.image_actor)
            self.image_actor = None
            self.image_resizer = None
            self.im_aspect = None

        if im_array is None:
            return

        self.im_aspect = float(im_array.shape[0]) / float(im_array.shape[1])
        self.image_actor,self.image_resizer = get_image_actor(im_array,actortype='vtkActor2D')

        self.renderer.AddActor(self.image_actor)
        self.on_resize()




class CalcamInteractorStyle2D(vtk.vtkInteractorStyleTerrain):
 
    def __init__(self,parent=None,newpick_callback=None,cursor_move_callback=None,focus_changed_callback=None,refresh_callback=None,pre_move_callback=None):
        # Set callbacks for all the controls
        self.AddObserver("LeftButtonPressEvent",self.on_left_click)
        self.AddObserver("MiddleButtonPressEvent",self.middle_press)
        self.AddObserver("MiddleButtonReleaseEvent",self.middle_release)
        self.AddObserver("MouseWheelForwardEvent",self.zoom_in)
        self.AddObserver("MouseWheelBackwardEvent",self.zoom_out)
        self.AddObserver("MouseMoveEvent",self.mouse_move)
        self.newpick_callback = newpick_callback
        self.cursor_move_callback = cursor_move_callback
        self.focus_changed_callback = focus_changed_callback
        self.refresh_callback = refresh_callback
        self.pre_move_callback = pre_move_callback
        self.allow_focus_change = True
        self.linked_interactors = []
        self.cursor_size = 0.03
        self.overlay_actors = []


    # Do various initial setup things, most of which can't be done at the time of __init__
    def init(self):

        # Get the interactor object
        self.interactor = self.GetInteractor()

        # Some other objects from higher up which I need access to
        self.vtkwindow = self.interactor.GetRenderWindow()
        renderers = self.vtkwindow.GetRenderers()
        renderers.InitTraversal()
        self.renderer = renderers.GetNextItemAsObject()

        self.im_dragging = False

        self.camera = self.renderer.GetActiveCamera()
        self.camera.ParallelProjectionOn()


        # Turn off any VTK responses to keyboard input (all necessary keyboard shortcuts etc are done in Qt)
        self.interactor.RemoveObservers('KeyPressEvent')
        self.interactor.RemoveObservers('CharEvent')

        # Add observer for catching window resizing
        self.vtkwindow.AddObserver("ModifiedEvent",self.on_resize)


        # Variables
        self.active_cursors = {}
        self.passive_cursors = {}
        self.next_cursor_id = 0

        self.focus_cursor = None

        self.image_actor = None
        self.overlay_actors = []

        self.overlay_alpha = 1


    def link_with(self,interactor):
        self.linked_interactors.append(interactor)

    def unlink_view(self):
        self.linked_interactors = []

    def sync_view(self):

        scale = self.camera.GetParallelScale()
        pos = self.camera.GetPosition()
        focalpoint = self.camera.GetFocalPoint()

        for interactor in self.linked_interactors:
            interactor.camera.SetParallelScale(scale)
            interactor.camera.SetPosition(pos)
            interactor.camera.SetFocalPoint(focalpoint)
            interactor.zoom_level = self.zoom_level
            interactor.update_cursor_style()

    # Use this image object
    def set_image(self,image,n_subviews=1,subview_lookup=lambda x,y: 0,hold_position=False):

        if self.image_actor is not None:

            # If removing the image entirely or changing image dimensions, remove any overlays
            if image is None or image.shape[:2] != self.image_actor.image.shape[:2]:
                for actor in self.overlay_actors:
                    self.renderer.RemoveActor(actor)
                self.overlay_actors = []

            # Remove existing image
            self.renderer.RemoveActor(self.image_actor)
            self.image_actor = None

        if image is not None:

            self.n_subviews = n_subviews
            self.subview_lookup = subview_lookup

            winsize = self.vtkwindow.GetSize()
            winaspect =  float(winsize[0])/float(winsize[1])

            self.image_actor = get_image_actor(image)

            self.renderer.AddActor2D(self.image_actor)

            bounds = self.image_actor.GetBounds()
            xc = bounds[0] + bounds[1] / 2
            yc = bounds[2] + bounds[3] / 2
            ye = bounds[3] - bounds[2]
            xe = bounds[1] - bounds[0]

            self.zoom_ref_cc = (xc,yc)

            im_aspect = xe / ye

            if winaspect >= im_aspect:
                # Base new zero size on y dimension
                self.zoom_ref_scale = 0.5*ye
            else:
                self.zoom_ref_scale = 0.5*xe/winaspect


            if not hold_position:
                # Reset view to fit the whole image on screen.
                self.zoom_level = 1.                
                self.camera.SetParallelScale(self.zoom_ref_scale)
                self.camera.SetPosition(xc,yc,1.)
                self.camera.SetFocalPoint(xc,yc,0.)


        # Re-shuffle cursors into different sub-views if needed.
        for cursor_id in self.active_cursors.keys():
            coords = self.get_cursor_coords(cursor_id)
            new_actor_list = [None] * self.n_subviews
            new_cursor3d_list = [None] * self.n_subviews
            for n_old in range(len(coords)):
                if coords[n_old] is not None:
                    new_subview = self.subview_lookup(coords[n_old][0],coords[n_old][1])
                    new_actor_list[new_subview] = self.active_cursors[cursor_id]['actors'][n_old]
                    self.renderer.RemoveActor(new_actor_list[new_subview])
                    new_cursor3d_list[new_subview] = self.active_cursors[cursor_id]['cursor3ds'][n_old]
            self.active_cursors[cursor_id] = {'cursor3ds':new_cursor3d_list,'actors':new_actor_list}

        # Finish re-adding active cursors and remove + re-add passive cursors (not sure why but
        # they disappear if I don't do this).
        for active_cursor in self.active_cursors.values():
            for actor in active_cursor['actors']:
                if actor is not None:
                    self.renderer.AddActor(actor)


        for passive_cursor in self.passive_cursors.values():
                self.renderer.RemoveActor(passive_cursor['actor'])
                self.renderer.AddActor(passive_cursor['actor'])           


        if self.refresh_callback is not None:
            self.refresh_callback()


    def get_image(self):
        if self.image_actor is None:
            return None
        else:
            return self.image_actor.image


    def set_subview_lookup(self,n_subviews,subview_lookup):
        self.n_subviews = n_subviews
        self.subview_lookup = subview_lookup

        # Re-shuffle cursors into different sub-views if needed.
        for cursor_id in self.active_cursors.keys():
            coords = self.get_cursor_coords(cursor_id)
            new_actor_list = [None] * self.n_subviews
            new_cursor3d_list = [None] * self.n_subviews
            for n_old in range(len(coords)):
                if coords[n_old] is not None:
                    new_subview = self.subview_lookup(coords[n_old][0],coords[n_old][1])
                    new_actor_list[new_subview] = self.active_cursors[cursor_id]['actors'][n_old]
                    self.renderer.RemoveActor(new_actor_list[new_subview])
                    new_cursor3d_list[new_subview] = self.active_cursors[cursor_id]['cursor3ds'][n_old]
            self.active_cursors[cursor_id] = {'cursor3ds':new_cursor3d_list,'actors':new_actor_list}

        # Finish re-adding active cursors
        for active_cursor in self.active_cursors.values():
            for actor in active_cursor['actors']:
                if actor is not None:
                    self.renderer.AddActor(actor)


    def get_n_cursors(self):
        return ( len(self.active_cursors) , len(self.passive_cursors) )

    # On the CAD view, middle click + drag to pan
    def middle_press(self,obj,event):
        self.im_dragging = True

    def middle_release(self,obj,event):
            self.im_dragging = False


    def set_overlay_image(self,overlay_image):

        if overlay_image is None:
            overlay_image = []
        elif not isinstance(overlay_image,list):
            overlay_image = [overlay_image]
        
        for actor in self.overlay_actors:
            self.renderer.RemoveActor(actor)
        self.overlay_actors = []

        for i,image in enumerate(overlay_image):
            scaling = np.array(self.image_actor.image.shape[:2]) / np.array(image.shape[:2])
            if scaling[0] != scaling[1]:
                raise ValueError('Provided overlay image is a different aspect ratio to the main image!')

            actor = get_image_actor(image,scaling=scaling[0])
            actor.SetPosition(0,0,0.01*(i+1))
            actor.GetProperty().SetOpacity(self.overlay_alpha)
            self.renderer.AddActor2D(actor)
            self.overlay_actors.append(actor)

        if self.refresh_callback is not None:
            self.refresh_callback()


    def set_overlay_alpha(self,alpha):

        self.overlay_alpha = alpha

        for actor in self.overlay_actors:
            actor.GetProperty().SetOpacity(alpha)

        if self.refresh_callback is not None:
            self.refresh_callback()

    # Left click to move a point or add a new point
    def on_left_click(self,obj,event):

        if self.image_actor is None:
            return

        clicked_cursor = None

        ctrl_pressed = self.interactor.GetControlKey()

        clickcoords = self.interactor.GetEventPosition()

        # Check if the click was near enough an existing cursor to be considered as clicking it
        dist = 7.
        for cid, cursor in self.active_cursors.items():

            if cid == self.focus_cursor:
                continue

            for icursor in cursor['cursor3ds']:
                if icursor is not None:
                    screencoords = self.image_to_screen_coords(icursor.GetFocalPoint())
                    dist_from_cursor = np.sqrt( (screencoords[0] - clickcoords[0])**2 + (screencoords[1] - clickcoords[1])**2 )
                    if dist_from_cursor < dist:
                            clicked_cursor = cid
                            dist = dist_from_cursor


        pickcoords = self.screen_to_image_coords(clickcoords)

        bounds = self.image_actor.GetBounds()
        maxind = np.array( [bounds[1] - bounds[0] - 1, bounds[3] - bounds[2] - 1] )

        if np.any(pickcoords < 0) or np.any(pickcoords > maxind):
            return


        if (ctrl_pressed or self.focus_cursor is None) and clicked_cursor is None:

            if self.newpick_callback is not None:
                self.newpick_callback(pickcoords)

        else:

            if clicked_cursor is not None and self.allow_focus_change:

                self.set_cursor_focus(clicked_cursor)

                if self.focus_changed_callback is not None:
                    self.focus_changed_callback(clicked_cursor)


            elif self.focus_cursor is not None:

                view_index = self.subview_lookup(pickcoords[0],pickcoords[1])

                if self.active_cursors[self.focus_cursor]['cursor3ds'][view_index] is None:
                    self.add_active_cursor(pickcoords,add_to=self.focus_cursor)
                else:
                    if self.pre_move_callback is not None:
                        self.pre_move_callback()
                    self.set_cursor_coords(self.focus_cursor,pickcoords,view_index)

                if self.cursor_move_callback is not None:
                    self.cursor_move_callback( self.focus_cursor, self.get_cursor_coords(self.focus_cursor))


        if self.refresh_callback is not None:
            self.refresh_callback()




    def zoom_in(self,obj,event):


        if self.image_actor is None:
            return

        winsize = self.vtkwindow.GetSize()

        # Re-position the image camera to keep the image point under the mouse pointer fixed when zooming
        zoom_ratio = 1 + 0.2/self.zoom_level                 # The zoom ratio we will have

        # Current camera position and scale
        campos = self.camera.GetPosition()
        camscale = self.camera.GetParallelScale() * 2.

        # Where is the zoom target in world coordinates?

        # Zoom coordinates in window pixels
        zoomcoords = list(self.interactor.GetEventPosition())

        # Position of current centre from to where we're zooming
        zoomcoords = ( (zoomcoords[0] - winsize[0]/2.)/winsize[0] * camscale * float(winsize[0])/float(winsize[1]) + campos[0],
                       (zoomcoords[1] - winsize[1]/2.)/winsize[1] * camscale + campos[1] )

        # Vector from zoom point to current camera centre
        zoomvec = ( campos[0] - zoomcoords[0] , campos[1] - zoomcoords[1] )

        # Now we move the camera along the line bwteen the current camera centre and zoom position
        newxc = zoomvec[0]/zoom_ratio + zoomcoords[0]
        newyc = zoomvec[1]/zoom_ratio + zoomcoords[1]

        # Actually move the camera
        self.camera.SetPosition((newxc,newyc,1.))
        self.camera.SetFocalPoint((newxc,newyc,0.))

        # Actually zoom in.
        self.zoom_level = self.zoom_level + 0.2
        self.camera.SetParallelScale(self.zoom_ref_scale / self.zoom_level)
        self.update_cursor_style()

        # Update any linked interactors
        self.sync_view()




    def zoom_out(self,obj,event):

        if self.image_actor is None:
            return

        # Zoom out smoothly until the whole image is visible
        if self.zoom_level > 1.:

            zoom_ratio = 0.2/(self.zoom_level**2 - 1.2*self.zoom_level + 0.2)

            campos = self.camera.GetPosition()

            zoomvec = ( self.zoom_ref_cc[0] - campos[0] , self.zoom_ref_cc[1] - campos[1] )

            self.camera.SetPosition((campos[0] + zoomvec[0] * zoom_ratio, campos[1] +  zoomvec[1] * zoom_ratio, 1.))
            self.camera.SetFocalPoint((campos[0] + zoomvec[0] * zoom_ratio, campos[1] +  zoomvec[1] * zoom_ratio, 0.))

            self.zoom_level = self.zoom_level - 0.2
            self.camera.SetParallelScale(self.zoom_ref_scale / self.zoom_level)
            self.update_cursor_style()

            self.sync_view()



    # Defocus cursors for a given point pair
    def set_cursor_focus(self,cursor_id):

        if cursor_id is not None:
            if cursor_id not in self.active_cursors.keys():
                raise ValueError('No cursor with ID {:d}'.format(cursor_id))

        self.focus_cursor = cursor_id
        self.update_cursor_style()


    def get_cursor_focus(self):
        return self.focus_cursor


    # Similar to Set3DCursorStyle but for image points
    def update_cursor_style(self):
        
        camscale = self.camera.GetParallelScale()

        for cid,cursor in self.active_cursors.items():

            for i,icursor in enumerate(cursor['cursor3ds']):
                if icursor is not None:

                    pos = icursor.GetFocalPoint()

                    if self.focus_cursor == cid:
                        colour = (0,0.8,0)
                        linewidth = 3
                        size = self.cursor_size * camscale
                    else:
                        colour = (0.8,0,0)
                        linewidth = 2
                        size = 0.5*self.cursor_size* camscale

                    icursor.SetModelBounds(pos[0]-size,pos[0]+size,pos[1]-size,pos[1]+size,0.0,0.0)
                    cursor['actors'][i].GetProperty().SetColor(colour)
                    cursor['actors'][i].GetProperty().SetLineWidth(linewidth)


        size = size = 0.015 * camscale
        for cursor in self.passive_cursors.values():
            pos = cursor['cursor3d'].GetFocalPoint()
            cursor['cursor3d'].SetModelBounds(pos[0]-size,pos[0]+size,pos[1]-size,pos[1]+size,0.,0.)

        if self.refresh_callback is not None:
            self.refresh_callback()


    # Adjust 2D image size and cursor positions if the window is resized
    def on_resize(self,obg=None,event=None):


        if self.image_actor is not None:

            vtksize = self.vtkwindow.GetSize()

            winaspect = float(vtksize[0])/float(vtksize[1])

            bounds = self.image_actor.GetBounds()
            ye = bounds[3] - bounds[2]
            xe = bounds[1] - bounds[0]

            im_aspect = xe / ye

            if winaspect >= im_aspect:
                # Base new zero size on y dimension
                self.zoom_ref_scale = 0.5*ye
            else:
                self.zoom_ref_scale = 0.5*xe/winaspect


            self.camera.SetParallelScale(self.zoom_ref_scale / self.zoom_level)

            self.sync_view()
        


    
    # Function to convert display coordinates to pixel coordinates on the camera image
    def screen_to_image_coords(self,screen_coords):

        vtksize = self.vtkwindow.GetSize()

        camyscale = self.camera.GetParallelScale() * 2.
        camxscale = camyscale * float(vtksize[0])/float(vtksize[1])
        cc = self.camera.GetFocalPoint()
        im_coords = [ (( screen_coords[0] - vtksize[0]/2. ) / vtksize[0]) * camxscale + cc[0], (screen_coords[1] - vtksize[1]/2.)/vtksize[1] * camyscale + cc[1]]

        bounds = self.image_actor.GetBounds()
        ysize = bounds[3] - bounds[2]

        im_coords[-1] = ysize - im_coords[-1]

        return np.array(im_coords)



    def image_to_screen_coords(self,image_coords):

        vtksize = self.vtkwindow.GetSize()

        camyscale = self.camera.GetParallelScale() * 2.
        camxscale = camyscale * float(vtksize[0])/float(vtksize[1])
        cc = self.camera.GetFocalPoint()

        screen_coords = ( float(image_coords[0] - cc[0]) / camxscale * vtksize[0] + vtksize[0]/2. , float(image_coords[1] - cc[1]) / camyscale*vtksize[1] + vtksize[1]/2.)

        return np.array(screen_coords)




    # Add a new point on the image
    def add_active_cursor(self,coords,add_to=None):

        subview = self.subview_lookup(coords[0],coords[1])

        if add_to is None:
            
            new_cursor_id = self.next_cursor_id
            self.next_cursor_id += 1

            self.active_cursors[new_cursor_id] = {'cursor3ds':[None]*self.n_subviews,'actors':[None]*self.n_subviews}
        else:
            new_cursor_id = add_to

        bounds = self.image_actor.GetBounds()
        ysize = bounds[3] - bounds[2]

        # Create new cursor and set it up
        new_cursor = vtk.vtkCursor3D()

        # Some setup of the cursor
        new_cursor.OutlineOff()
        new_cursor.XShadowsOff()
        new_cursor.YShadowsOff()
        new_cursor.ZShadowsOff()
        new_cursor.AxesOn()
        new_cursor.TranslationModeOn()

        new_cursor.SetFocalPoint([coords[0],ysize-coords[1],0.05])

        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputConnection(new_cursor.GetOutputPort())

        actor = vtk.vtkActor()
        actor.SetMapper(mapper)

        self.active_cursors[new_cursor_id]['actors'][subview] = actor
        self.active_cursors[new_cursor_id]['cursor3ds'][subview] = new_cursor


        # Add new cursor to screen
        self.renderer.AddActor(actor)

        self.update_cursor_style()

        if self.refresh_callback is not None:
            self.refresh_callback()

        return new_cursor_id



    def remove_active_cursor(self,cursor_id):

        if cursor_id not in self.active_cursors:
            raise ValueError('No such cursor ID {:d} exists!'.format(cursor_id))

        cursor = self.active_cursors.pop(cursor_id)

        for actor in cursor['actors']:
            self.renderer.RemoveActor(actor)

        if self.focus_cursor == cursor_id:
            self.focus_cursor = None

        if self.refresh_callback is not None:
            self.refresh_callback()


    def get_cursor_coords(self,cursor_id):

        if cursor_id not in self.active_cursors:
            raise ValueError('No such cursor ID {:d} exists!'.format(cursor_id))        

        coords = []

        bounds = self.image_actor.GetBounds()
        ysize = bounds[3] - bounds[2]

        for cursor in self.active_cursors[cursor_id]['cursor3ds']:
            if cursor is not None:
                icoords = np.array(cursor.GetFocalPoint()[:2])
                icoords[-1] = ysize - icoords[-1]
                coords.append(icoords)
            else:
                coords.append(None)

        return coords


    def set_cursor_coords(self,cursor_id,coords,subview=None):

        if cursor_id not in self.active_cursors:
            raise ValueError('No such cursor ID {:d} exists!'.format(cursor_id))   

        if subview is None and len(coords) != len(self.active_cursors[cursor_id]['cursor3ds']):
            raise ValueError('Expected {:d} sets of coordinates; instead got {:d}.'.format(len(self.active_cursors[cursor_id]['cursor3ds']),len(coords)))

        bounds = self.image_actor.GetBounds()
        ysize = bounds[3] - bounds[2]

        if subview is None:
            for i,cursor in enumerate(self.active_cursors[cursor_id]['cursor3ds']):
                cursor.SetFocalPoint(coords[i][0],ysize-coords[i][1],0.05)
        else:
            self.active_cursors[cursor_id]['cursor3ds'][subview].SetFocalPoint(coords[0],ysize-coords[1],0.05)


    # Show the current CAD points re-projected on to the image
    # using the current fit.
    def add_passive_cursor(self,coords):
            
        new_cursor_id = self.next_cursor_id
        self.next_cursor_id += 1

        bounds = self.image_actor.GetBounds()
        ysize = bounds[3] - bounds[2]

        # Create new cursor and set it up
        new_cursor = vtk.vtkCursor3D()

        # Some setup of the cursor
        new_cursor.OutlineOff()
        new_cursor.XShadowsOff()
        new_cursor.YShadowsOff()
        new_cursor.ZShadowsOff()
        new_cursor.AxesOn()
        new_cursor.TranslationModeOn()

        new_cursor.SetFocalPoint([coords[0],ysize-coords[1],0.05])

        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputConnection(new_cursor.GetOutputPort())

        actor = vtk.vtkActor()
        actor.SetMapper(mapper)
        actor.GetProperty().SetColor((0,0,1.))
        actor.GetProperty().SetLineWidth(2)

        self.passive_cursors[new_cursor_id] = {'cursor3d':new_cursor,'actor': actor}

        # Add new cursor to screen
        self.renderer.AddActor(actor)

        self.update_cursor_style()

        if self.refresh_callback is not None:
            self.refresh_callback()

        return new_cursor_id


    def remove_passive_cursor(self,cursor_id):

        if cursor_id not in self.passive_cursors:
            raise ValueError('No such cursor ID {:d} exists!'.format(cursor_id))

        cursor = self.passive_cursors.pop(cursor_id)

        self.renderer.RemoveActor(cursor['actor'])

        if self.refresh_callback is not None:
            self.refresh_callback()


    def clear_passive_cursors(self):

        for cursor in self.passive_cursors.values():
            self.renderer.RemoveActor(cursor['actor'])

        self.passive_cursors = {}

        if self.refresh_callback is not None:
            self.refresh_callback()



    # Custom mouse move event to enable middle click panning on both
    # CAD and image views.
    def mouse_move(self,obj,event):

        if self.im_dragging and self.zoom_level > 1:

            winsize = self.vtkwindow.GetSize()

            lastXYpos = self.interactor.GetLastEventPosition() 
            xypos = self.interactor.GetEventPosition()
            camscale = self.camera.GetParallelScale() * 2
            oldpos = self.camera.GetPosition()
            deltaX = (xypos[0] - lastXYpos[0])/float(winsize[0]) * camscale * float(winsize[0])/float(winsize[1])
            deltaY = (xypos[1] - lastXYpos[1])/float(winsize[1]) * camscale

            newY = oldpos[1] - deltaY
            newX = oldpos[0] - deltaX


            # Make sure we don't pan outside the image.
            im_bounds = self.image_actor.GetBounds()
            xcamscale = camscale * float(winsize[0])/winsize[1]
            if newX + xcamscale/2. > im_bounds[1] and newX - xcamscale/2. > im_bounds[0] and deltaX < 0:
                newX = oldpos[0]
            elif newX - xcamscale/2. < im_bounds[0] and newX + xcamscale/2. < im_bounds[1] and deltaX > 0:
                newX = oldpos[0]
            if newY + camscale/2. > im_bounds[3] and newY - camscale/2. > im_bounds[2] and deltaY < 0:
                newY = oldpos[1]
            elif newY - camscale/2. < im_bounds[2] and newY + camscale/2. < im_bounds[3] and deltaY > 0:
                newY = oldpos[1]

            # Move image camera
            self.camera.SetPosition((newX, newY,1.))
            self.camera.SetFocalPoint((newX,newY,0.))

            self.sync_view()

            if self.refresh_callback is not None:
                self.refresh_callback()


'''
* Copyright 2015-2022 European Atomic Energy Community (EURATOM)
*
* Licensed under the EUPL, Version 1.1 or - as soon they
  will be approved by the European Commission - subsequent
  versions of the EUPL (the "Licence");
* You may not use this work except in compliance with the
  Licence.
* You may obtain a copy of the Licence at:
*
* https://joinup.ec.europa.eu/software/page/eupl
*
* Unless required by applicable law or agreed to in
  writing, software distributed under the Licence is
  distributed on an "AS IS" basis,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
  express or implied.
* See the Licence for the specific language governing
  permissions and limitations under the Licence.
'''


import os
import json
import sys
import glob
import traceback
import multiprocessing
import warnings

from .io import ZipSaveFile
from .misc import import_source, unload_source


# Number of CPUs to use for multiprocessing enabled
# calculations. Always initialised to (n_cpus - 1)
# and any changes by the user only apply to that session.
n_cpus = max(1,multiprocessing.cpu_count()-1)

# Path where the "built in" image source code lives; this is always in the calcam source directory.
builtin_imsource_path = os.path.join(os.path.split(os.path.abspath(__file__))[0],'builtin_image_sources')

# File where calcam stores the user's configuration
user_cfg_path = os.path.expanduser('~/.calcam_config')

# If the user doesn't have their own configuration, look for a default config file in the calcam install directory.
# This can be used e.g. if installing calcam on a multi-user system where you want to provide a default config for all users.
# But there is not one included by default with Calcam!
default_cfg_path = os.path.join(os.path.split(os.path.abspath(__file__))[0],'site_defaults.cfg')

# Filename filters for different types of file
filename_filters = {'calibration':'Calcam Calibration (*.ccc)','image':'PNG Image (*.png)','pointpairs':'Calcam Point Pairs (*.ccc *.csv)','movement':'Calcam Affine Transform (*.cmc)'}

class CalcamConfig():
    '''
    Class to represent the persistent calcam settings.
    
    It is convenient to use a class for this because whenever this is
    instantiated somewhere in Calcam, we always get the most recent configuration
    from disk.
    '''
    def __init__(self):

        # Configuration fields and their default values
        self.fields = {'image_source_paths': [],
                       'cad_def_paths': [],
                       'file_dirs': {},
                       'default_model': None,
                       'default_image_source':'Image File',
                       'mouse_sensitivity':75,
                       'main_overlay_colour':(0,0,1.,0.6),
                       'second_overlay_colour':(1.,0,0,0.6)
                       }

        # Filename filters (which should never need to change so are defined above)
        self.filename_filters = filename_filters

        # Try to load the user's own config
        try:
            with open(user_cfg_path,'r') as f:
                user_dict = json.load(f)
        except IOError:
            user_dict = {}

        # Try to load the default config
        try:
            with open(default_cfg_path, 'r') as f:
                defaults_dict = json.load(f)
        except IOError:
            defaults_dict = {}

        # For each field, first try to get the user's own value,
        # then the one from the defaults file, then fall back to the version defined above.
        for key in self.fields:
            if key in user_dict:
                setattr(self,key,user_dict[key])
            elif key in defaults_dict:
                setattr(self,key,defaults_dict[key])
            else:
                setattr(self,key,self.fields[key])

        self.save()


    def save(self):
        """
        Save the configuration to disk in the user's home directory.
        """
        save_dict = { key : getattr(self,key) for key in self.fields}

        try:
            with open(user_cfg_path,'w') as f:
                json.dump(save_dict,f,indent=4)
        except Exception as e:
            warnings.warn('Could not save user calcam configuration to file "{:s}": {:}'.format(user_cfg_path,e))



    def get_cadmodels(self):
        """
        Get the list of available CAD models.

        Returns:
            A dictionary where the keys are human-readable names of the CAD model and the velues are a list of: \
            [Filename of the CAD model, List of features in that CAD model, Default variant name]
        """
        cadmodels = {}

        for path in self.cad_def_paths:
            filelist = glob.glob(os.path.join(path,'*.ccm'))

            for fname in filelist:

                try:
                    with ZipSaveFile(fname,'rs') as f:
                        with f.open_file('model.json','r') as j: 
                               caddef = json.load(j)
                except:
                    continue

                if caddef['machine_name'] not in cadmodels:
                    key = caddef['machine_name']
                else:
                    
                    existing_model = cadmodels.pop(caddef['machine_name'])
                    existing_key = '{:s} [{:s}/{:s}]'.format(caddef['machine_name'], existing_model[0].split(os.sep)[-2],os.path.split(existing_model[0])[-1] )
                    cadmodels[existing_key] = existing_model

                    key = '{:s} [{:s}/{:s}]'.format(caddef['machine_name'], fname.split(os.sep)[-2],os.path.split(fname)[1] )

                cadmodels[key] = [fname,[str(x) for x in caddef['features'].keys()],caddef['default_variant']]

        return cadmodels


    def get_image_sources(self,meta_only=False):
        """
        Get a list of available image source providers.
        """
        image_sources = []
        displaynames = []
        meta = []

        for path in [builtin_imsource_path] + self.image_source_paths:

            filelist = glob.glob(os.path.join(path,'*'))

            trylist = []
            for f in filelist:
                if (os.path.isdir(f) and os.path.isfile(os.path.join(f,'__init__.py'))) or f.endswith('.py'):
                    trylist.append(f)

            for fname in trylist:
                if fname.endswith('__init__.py'):
                    tidy_name = os.sep.join(fname.split(os.sep)[-3:-1])
                    fname = os.path.split(fname)[0]
                else:
                    tidy_name = os.sep.join(fname.split(os.sep)[-2:])

                try:
                    # Import the module, check it has the right attributes and add its info to the metadata table
                    usermodule = import_source(fname)

                    try:
                        if not callable(usermodule.get_image_function):
                            raise ImportError()
                    except Exception:
                        meta.append([tidy_name, fname, 'Not a valid image source definition:\nDoes not contain required function "get_image_function(..)"'])
                        continue

                    try:
                        if type(usermodule.get_image_arguments) is not list:
                            raise ImportError
                    except Exception:
                        meta.append([tidy_name, fname, 'Not a valid image source definition:\nDoes not contain required list attribute "get_image_arguments"'])
                        continue

                    try:
                        if type(usermodule.display_name) is not str:
                            raise ImportError
                    except Exception:
                        meta.append([tidy_name, fname, 'Not a valid image source definition:\nDoes not contain required string attribute  "display_name"'])
                        continue


                    # Make sure the imported modules have unique display names
                    if usermodule.display_name in displaynames:
                        old_ind = displaynames.index(usermodule.display_name)
                        for i,metadata in enumerate(meta):
                            if metadata[0] == usermodule.display_name:
                                old_meta_ind = i
                                break

                        usermodule.display_name = usermodule.display_name + ' [{:s}]'.format(tidy_name)
                        other_path = image_sources[old_ind].__file__
                        if other_path.endswith('__init__.py'):
                            tidyname = os.sep.join(other_path.split(os.sep)[-3:-1])
                        else:
                            tidyname = os.sep.join(other_path.split(os.sep)[-2:])
                        image_sources[old_ind].display_name = image_sources[old_ind].display_name + ' [{:s}]'.format(tidyname)
                        meta[old_meta_ind][0] = image_sources[old_ind].display_name

                    displaynames.append(usermodule.display_name)
                    image_sources.append(usermodule)
                    meta.append([usermodule.display_name,fname,None])
                except Exception:
                    # If it won't import or doesn't have the right attributes, show this in the metadata
                    tb_info = ''.join(traceback.format_exception(*sys.exc_info(), limit=-1))
                    meta.append([tidy_name,fname,'Cannot be imported:\n{:s}'.format(tb_info)])
                    continue

                # Built-in image sources get special metadata
                if path == builtin_imsource_path:
                    meta[-1][1] = None


        if meta_only:
            for module_meta in meta:
                if module_meta[-1] is None and module_meta[1] is not None:
                    unload_source(module_meta[1])
            return meta
        else:
            return image_sources



'''
* Copyright 2015-2018 European Atomic Energy Community (EURATOM)
*
* Licensed under the EUPL, Version 1.1 or - as soon they
  will be approved by the European Commission - subsequent
  versions of the EUPL (the "Licence");
* You may not use this work except in compliance with the
  Licence.
* You may obtain a copy of the Licence at:
*
* https://joinup.ec.europa.eu/software/page/eupl
*
* Unless required by applicable law or agreed to in
  writing, software distributed under the Licence is
  distributed on an "AS IS" basis,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
  express or implied.
* See the Licence for the specific language governing
  permissions and limitations under the Licence.
'''


import copy

import matplotlib.cm

from .core import *
from .vtkinteractorstyles import CalcamInteractorStyle2D, CalcamInteractorStyle3D
from ..render import render_cam_view,get_wall_coverage_actor,render_hires
from ..coordtransformer import CoordTransformer
from ..raycast import raycast_sightlines
from ..image_enhancement import enhance_image, scale_to_8bit
from ..movement import manual_movement
from ..calibration import NoSubviews

type_description = {'alignment': 'Manual Alignment', 'fit':'Point pair fitting','virtual':'Virtual'}

cmaps = []
for cmap in [('Grayscale','gray'),('Inferno','inferno'),('Magma','magma'),('Plasma','plasma'),('Viridis','viridis'),('JET','jet'),('HSV','hsv')]:
    try:
        cmaps.append((cmap[0],matplotlib.cm.get_cmap(cmap[1])))
    except Exception:
        continue

# Main calcam window class for actually creating calibrations.
class ImageAnalyser(CalcamGUIWindow):

    def __init__(self, app, parent = None):

        # GUI initialisation
        CalcamGUIWindow.init(self,'image_analyser.ui',app,parent)

        # Start up with no CAD model
        self.cadmodel = None
        self.calibration = None

        # Set up VTK
        self.qvtkwidget_3d = qt.QVTKRenderWindowInteractor(self.vtkframe_3d)
        self.vtkframe_3d.layout().addWidget(self.qvtkwidget_3d)
        self.interactor3d = CalcamInteractorStyle3D(refresh_callback=self.refresh_3d,viewport_callback=self.update_viewport_info,resize_callback=self.update_vtk_size,newpick_callback=self.update_from_3d,cursor_move_callback=lambda cid,coords: self.update_from_3d(coords))
        self.interactor3d.allow_focus_change = False
        self.qvtkwidget_3d.SetInteractorStyle(self.interactor3d)
        self.renderer_3d = vtk.vtkRenderer()
        self.renderer_3d.SetBackground(0, 0, 0)
        self.qvtkwidget_3d.GetRenderWindow().AddRenderer(self.renderer_3d)
        self.camera_3d = self.renderer_3d.GetActiveCamera()

        self.qvtkwidget_2d = qt.QVTKRenderWindowInteractor(self.vtkframe_2d)
        self.vtkframe_2d.layout().addWidget(self.qvtkwidget_2d)
        self.interactor2d = CalcamInteractorStyle2D(refresh_callback=self.refresh_2d,newpick_callback=lambda x: self.update_from_2d([x]),cursor_move_callback=lambda cid,coords: self.update_from_2d(coords))
        self.interactor2d.allow_focus_change = False
        self.qvtkwidget_2d.SetInteractorStyle(self.interactor2d)
        self.renderer_2d = vtk.vtkRenderer()
        self.renderer_2d.SetBackground(0, 0, 0)
        self.qvtkwidget_2d.GetRenderWindow().AddRenderer(self.renderer_2d)
        self.camera_2d = self.renderer_2d.GetActiveCamera()

        self.image_geometry = None
        self.image_assumptions = []
        self.image_load_coords = None
        self.mov_correction = None
        self.im_projection = None

        self.coords_3d = None

        self.populate_models()



        # Disable image transform buttons if we have no image
        self.image_settings.hide()

        self.tabWidget.setTabEnabled(2,False)
        self.tabWidget.setTabEnabled(3,False)

        self.cmap_dropdown.addItems([cmap[0] for cmap in cmaps])
        self.cmap_dropdown.setCurrentIndex(0)

        # Callbacks for GUI elements
        self.image_sources_list.currentIndexChanged.connect(self.build_imload_gui)
        self.viewlist.itemSelectionChanged.connect(self.change_cad_view)
        self.camX.valueChanged.connect(self.change_cad_view)
        self.camY.valueChanged.connect(self.change_cad_view)
        self.camZ.valueChanged.connect(self.change_cad_view)
        self.tarX.valueChanged.connect(self.change_cad_view)
        self.tarY.valueChanged.connect(self.change_cad_view)
        self.tarZ.valueChanged.connect(self.change_cad_view)
        self.cam_roll.valueChanged.connect(self.change_cad_view)
        self.camFOV.valueChanged.connect(self.change_cad_view)
        self.load_model_button.clicked.connect(self.load_model)
        self.model_name.currentIndexChanged.connect(self.populate_model_variants)
        self.feature_tree.itemChanged.connect(self.update_checked_features)
        self.load_image_button.clicked.connect(self.load_image)
        self.load_calib_button.clicked.connect(self.load_calib)
        self.reset_view_button.clicked.connect(lambda: self.set_view_from_calib(self.calibration,0))
        self.cursor_closeup_button.clicked.connect(self.set_view_to_cursor)
        self.cal_props_button.clicked.connect(self.show_calib_info)
        self.viewport_load_calib.clicked.connect(self.load_viewport_calib)
        self.cmap_checkbox.toggled.connect(self.update_cmap)
        self.cmap_dropdown.currentIndexChanged.connect(self.update_cmap)
        self.cmap_min.valueChanged.connect(self.update_cmap)
        self.cmap_max.valueChanged.connect(self.update_cmap)
        self.movement_correction_button.clicked.connect(self.update_movement_correction)
        self.im_mapping_checkbox.toggled.connect(self.update_mapped_image)
        self.mapped_im_opacity_slider.valueChanged.connect(self.update_mapped_image)
        self.render_button.clicked.connect(self.save_image)
        self.render_image_checkbox.toggled.connect(self.update_render_resolution)
        self.render_cad_checkbox.toggled.connect(self.update_render_resolution)
        self.overlay_opacity_slider.valueChanged.connect(self.change_overlay_colour)
        self.overlay_type_box.currentIndexChanged.connect(self.update_overlay)
        self.overlay_colour_button.clicked.connect(self.change_overlay_colour)

        self.sightline_checkbox.toggled.connect(lambda: self.update_from_3d(self.coords_3d))

        self.overlay_checkbox.toggled.connect(self.update_overlay)

        self.enhance_checkbox.stateChanged.connect(self.toggle_enhancement)

        self.image = None
        self.cmap_options.hide()
        self.viewport_calibs = DodgyDict()

        self.mapped_im_opacity_slider.hide()
        self.mapped_im_opacity_label.hide()

        self.control_sensitivity_slider.valueChanged.connect(lambda x: self.interactor3d.set_control_sensitivity(x*0.01))
        self.rmb_rotate.toggled.connect(self.interactor3d.set_rmb_rotate)
        self.interactor3d.set_control_sensitivity(self.control_sensitivity_slider.value()*0.01)


        # Populate image sources list and tweak GUI layout for image loading.
        self.imload_inputs = {}
        self.image_load_options.layout().setColumnMinimumWidth(0,100)

        self.image_sources = self.config.get_image_sources()
        index = -1
        for i,imsource in enumerate(self.image_sources):
            self.image_sources_list.addItem(imsource.display_name)
            if imsource.display_name == self.config.default_image_source:
                index = i

        self.image_sources_list.setCurrentIndex(index)

        self.cursor_ids = {'3d':None,'2d':{'visible':None,'hidden':None}}
        self.sightline_actors = []
        self.overlay_colour = [0,0,1,0.5]

        self.overlay_settings.hide()

        # Start the GUI!
        self.show()
        self.interactor2d.init()
        self.interactor3d.init()
        self.qvtkwidget_3d.GetRenderWindow().GetInteractor().Initialize()
        self.qvtkwidget_2d.GetRenderWindow().GetInteractor().Initialize()
        self.interactor3d.on_resize()


    def update_render_resolution(self):

        w = self.vtksize[0]
        h = self.vtksize[1]

        if not self.render_image_checkbox.isChecked() and not self.render_cad_checkbox.isChecked():
            self.render_button.setEnabled(False)
            self.render_resolution.setEnabled(False)
        else:
            self.render_button.setEnabled(True)
            self.render_resolution.setEnabled(True)

        if self.render_image_checkbox.isChecked():
            self.render_resolution.setCurrentIndex(0)
            self.render_resolution.setEnabled(False)
        else:
            self.render_resolution.setEnabled(True)

        if self.render_image_checkbox.isChecked() and self.render_cad_checkbox.isChecked():
            w = w * 2

        index = max(self.render_resolution.currentIndex(), 0)
        self.render_resolution.clear()
        self.render_resolution.addItem('{:d} x {:d} (Window Size)'.format(w, h))
        self.render_resolution.addItem('{:d} x {:d}'.format(w * 2, h * 2))
        self.render_resolution.addItem('{:d} x {:d}'.format(w * 4, h * 4))
        self.render_resolution.setCurrentIndex(index)



    def update_vtk_size(self,vtksize):

        self.vtksize = vtksize
        self.update_render_resolution()


    def save_image(self):

        filename = self.get_save_filename('image')
        if filename is None:
            return

        self.app.setOverrideCursor(qt.QCursor(qt.Qt.WaitCursor))
        self.statusbar.showMessage('Rendering image to {:s} ...'.format(filename))

        c3d = None
        c2d = []
        if not self.render_cursor_checkbox.isChecked():
            if self.interactor3d.focus_cursor is not None:
                c3d = self.interactor3d.cursors[self.interactor3d.focus_cursor]['actor']
            if self.interactor2d.focus_cursor is not None:
                for actor in self.interactor2d.active_cursors[self.interactor2d.focus_cursor]['actors']:
                    c2d.append(actor)

            self.renderer_3d.RemoveActor(c3d)
            for actor in c2d:
                self.renderer_2d.RemoveActor(actor)

        oversampling = 2 ** (self.render_resolution.currentIndex())

        h = self.vtksize[1] * oversampling
        im = np.zeros((h,0,3),dtype=np.uint8)

        if self.render_image_checkbox.isChecked():
            im = np.hstack((im,render_hires(self.renderer_2d,oversampling=oversampling)))
        if self.render_cad_checkbox.isChecked():
            im = np.hstack((im,render_hires(self.renderer_3d,oversampling=oversampling)))

        im[:,:,:3] = im[:,:,2::-1]
        cv2.imwrite(filename,im)

        if c3d is not None:
            self.renderer_3d.AddActor(c3d)
        for actor in c2d:
            self.renderer_2d.AddActor(actor)

        self.app.restoreOverrideCursor()
        self.statusbar.clearMessage()

    def update_movement_correction(self):

        if self.calibration is not None:
            transformer = self.calibration.geometry
        else:
            transformer = self.image_geometry

        self.mov_correction = manual_movement(self.calibration,transformer.original_to_display_image(self.image),self.mov_correction,parent_window=self)

        if self.coords_3d is not None:
            self.update_from_3d(self.coords_3d)

        if self.overlay_checkbox.isChecked():
            self.update_overlay()

        self.update_mapped_image()


    def update_from_3d(self,coords_3d):

        # Sight line drawing style
        visible_linewidth = 3
        visible_colour = (0,0.8,0)
        invisible_linewidth = 1
        invisible_colour = (0.8,0,0)

        if self.calibration is not None and self.image is not None and coords_3d is not None:

            self.cursor_closeup_button.setEnabled(True)

            # Clear anything which already exists
            if self.cursor_ids['3d'] is not None:
                self.interactor3d.remove_cursor(self.cursor_ids['3d'])
                self.cursor_ids['3d'] = None

            if self.cursor_ids['2d']['visible'] is not None:
                self.interactor2d.remove_active_cursor(self.cursor_ids['2d']['visible'])
                self.cursor_ids['2d']['visible'] = None

            if self.cursor_ids['2d']['hidden'] is not None:
                self.interactor2d.remove_active_cursor(self.cursor_ids['2d']['hidden'])
                self.cursor_ids['2d']['hidden'] = None

            for actor in self.sightline_actors:
                self.interactor3d.remove_extra_actor(actor)
            self.sightline_actors = []


            intersection_coords = None

            visible = [False] * self.calibration.n_subviews

            # Find where the cursor(s) is/are in 2D.
            image_pos_nocheck = self.calibration.project_points([coords_3d],coords='original')

            image_pos = self.calibration.project_points([coords_3d],check_occlusion_with=self.cadmodel,occlusion_tol=1e-2,coords='original')

            for i in range(len(image_pos)):
                image_pos[i][0][:] = self.calibration.geometry.original_to_display_coords(*image_pos[i][0])
                image_pos_nocheck[i][0][:] = self.calibration.geometry.original_to_display_coords(*image_pos_nocheck[i][0])

                if np.any(np.isnan(image_pos_nocheck[i])):
                    visible[i] = False
                    continue

                raydata = raycast_sightlines(self.calibration,self.cadmodel,image_pos_nocheck[i][0,0],image_pos_nocheck[i][0,1],verbose=False,force_subview=i)
                visible[i] =True

                if self.mov_correction is not None:
                    image_pos_nocheck[i][0][:] = self.mov_correction.ref_to_moved_coords(*image_pos_nocheck[i][0])

                if np.any(np.isnan(image_pos[i])):
                    visible[i] = False
                    intersection_coords = raydata.ray_end_coords

                if visible[i]:

                    if self.cursor_ids['2d']['visible'] is None:
                        self.cursor_ids['2d']['visible'] = self.interactor2d.add_active_cursor(image_pos_nocheck[i][0,:])
                    else:
                        self.interactor2d.add_active_cursor(image_pos_nocheck[i][0,:],add_to=self.cursor_ids['2d']['visible'])

                    if self.sightline_checkbox.isChecked():
                        linesource = vtk.vtkLineSource()
                        pupilpos = self.calibration.get_pupilpos(subview=i)
                        linesource.SetPoint1(pupilpos[0],pupilpos[1],pupilpos[2])
                        linesource.SetPoint2(coords_3d)
                        mapper = vtk.vtkPolyDataMapper()
                        mapper.SetInputConnection(linesource.GetOutputPort())
                        actor = vtk.vtkActor()
                        actor.SetMapper(mapper)
                        actor.GetProperty().SetLineWidth(visible_linewidth)
                        actor.GetProperty().SetColor(visible_colour)
                        self.sightline_actors.append(actor)
                        self.interactor3d.add_extra_actor(actor)

                        model_extent = self.cadmodel.get_extent()
                        model_size = model_extent[1::2] - model_extent[::2]
                        max_ray_length = model_size.max() * 4
                        sldir = coords_3d - pupilpos
                        sldir = sldir / np.sqrt(np.sum(sldir**2))
                        rayend = pupilpos + sldir*max_ray_length

                        linesource = vtk.vtkLineSource()
                        linesource.SetPoint1(coords_3d)
                        linesource.SetPoint2(rayend)
                        mapper = vtk.vtkPolyDataMapper()
                        mapper.SetInputConnection(linesource.GetOutputPort())
                        actor = vtk.vtkActor()
                        actor.SetMapper(mapper)
                        actor.GetProperty().SetLineWidth(invisible_linewidth)
                        actor.GetProperty().SetColor(invisible_colour)
                        self.sightline_actors.append(actor)
                        self.interactor3d.add_extra_actor(actor)



                else:
                    if self.cursor_ids['2d']['hidden'] is None:
                        self.cursor_ids['2d']['hidden'] = self.interactor2d.add_active_cursor(image_pos_nocheck[i][0,:])
                    else:
                        self.interactor2d.add_active_cursor(image_pos_nocheck[i][0,:],add_to=self.cursor_ids['2d']['hidden'])

                    if self.sightline_checkbox.isChecked():
                        linesource = vtk.vtkLineSource()
                        pupilpos = self.calibration.get_pupilpos(subview=i)
                        linesource.SetPoint1(*pupilpos)
                        linesource.SetPoint2(*intersection_coords)
                        mapper = vtk.vtkPolyDataMapper()
                        mapper.SetInputConnection(linesource.GetOutputPort())
                        actor = vtk.vtkActor()
                        actor.SetMapper(mapper)
                        actor.GetProperty().SetLineWidth(invisible_linewidth)
                        actor.GetProperty().SetColor(visible_colour)
                        self.sightline_actors.append(actor)
                        self.interactor3d.add_extra_actor(actor)

                        linesource = vtk.vtkLineSource()
                        linesource.SetPoint1(*intersection_coords)
                        linesource.SetPoint2(*coords_3d)
                        mapper = vtk.vtkPolyDataMapper()
                        mapper.SetInputConnection(linesource.GetOutputPort())
                        actor = vtk.vtkActor()
                        actor.SetMapper(mapper)
                        actor.GetProperty().SetLineWidth(invisible_linewidth)
                        actor.GetProperty().SetColor(invisible_colour)
                        self.sightline_actors.append(actor)
                        self.interactor3d.add_extra_actor(actor)

            self.cursor_ids['3d'] = self.interactor3d.add_cursor(coords_3d)

            if self.cursor_ids['2d']['visible'] is not None:
                self.interactor2d.set_cursor_focus(self.cursor_ids['2d']['visible'])
                self.interactor3d.set_cursor_focus(self.cursor_ids['3d'])


            self.coords_2d = image_pos_nocheck
            self.coords_3d = coords_3d

            self.update_position_info(self.coords_2d,self.coords_3d,visible)




    def update_from_2d(self,coords_2d):

        if self.calibration is not None and self.image is not None and self.cadmodel is not None:

            for i,coords in enumerate(coords_2d):
                if coords is not None:
                    if self.mov_correction is not None:
                        coords = self.mov_correction.moved_to_ref_coords(*coords)

                    if self.calibration.subview_lookup(*coords) == -1:
                        raise UserWarning('The clicked position is outside the calibrated field of view.')
                    elif np.any(coords != self.coords_2d[i]):
                        raydata = raycast_sightlines(self.calibration,self.cadmodel,coords[0],coords[1],coords='Display',verbose=False)
                        self.update_from_3d(raydata.ray_end_coords[0,:])
                        return


    def on_close(self):
        self.qvtkwidget_3d.close()
        self.qvtkwidget_2d.close()


    def update_mapped_image(self):

        if self.sender() is self.mapped_im_opacity_slider:
            if self.im_projection is not None:
                self.im_projection.GetProperty().SetOpacity(self.mapped_im_opacity_slider.value()/100)
            self.refresh_3d()
            return

        if self.im_projection is not None:
            self.renderer_3d.RemoveActor(self.im_projection)

        if self.im_mapping_checkbox.isChecked():
            self.app.setOverrideCursor(qt.QCursor(qt.Qt.WaitCursor))
            self.statusbar.showMessage('Mapping image to CAD surface...')
            self.app.processEvents()

            if self.mov_correction is not None:
                im = self.mov_correction.warp_moved_to_ref(self.interactor2d.get_image())[0]
            else:
                im = self.interactor2d.get_image()

            if self.im_projection is not None:
                try:
                    self.im_projection.update_image(im)
                except ValueError:
                    actor = get_wall_coverage_actor(self.calibration, cadmodel=self.cadmodel, image=im,imagecoords='Display')
                    actor.GetProperty().SetOpacity(self.mapped_im_opacity_slider.value() / 100)
                    self.im_projection = actor
            else:
                actor = get_wall_coverage_actor(self.calibration,cadmodel=self.cadmodel,image=im,imagecoords='Display')
                actor.GetProperty().SetOpacity(self.mapped_im_opacity_slider.value() / 100)
                self.im_projection = actor

            self.renderer_3d.AddActor(self.im_projection)
            self.mapped_im_opacity_slider.show()
            self.mapped_im_opacity_label.show()
            self.app.restoreOverrideCursor()
            self.statusbar.clearMessage()

        else:
            self.mapped_im_opacity_slider.hide()
            self.mapped_im_opacity_label.hide()

        self.refresh_3d()

    def load_calib(self):

        opened_calib = self.object_from_file('calibration')

        # If we get no calibration from the above call it means the user has cancelled
        if opened_calib is None:
            return

        if None in opened_calib.view_models:
            raise UserWarning('The selected calibration file does not contain a full set of calibration parameters. Only calibration files containing all calibration parameters can be used.')

        self.mov_correction = None

        if self.image is not None:
            if self.image_load_coords.lower() == 'display':

                if self.image_geometry.get_display_shape() != opened_calib.geometry.get_display_shape() or (self.image_geometry.offset != opened_calib.geometry.offset and 'image_offset' not in self.image_assumptions):
                    osize = opened_calib.geometry.display_to_original_shape(self.image_geometry.get_display_shape())
                    try:
                        opened_calib.set_detector_window((self.image_geometry.offset[0],self.image_geometry.offset[1],osize[0],osize[1]))
                    except NoSubviews:
                        raise UserWarning('The detector area according to the image metadata ({:d}x{:d} pixels at offset {:d}x{:d}) does not contain any image according to the calibration! Perhaps the image detector offset is set incorrectly?'.format(osize[0],osize[1],self.image_geometry.offset[0],self.image_geometry.offset[1]))
                if self.image_geometry.transform_actions == []:
                    self.image = opened_calib.geometry.display_to_original_image(self.image)

            else:
                if self.image_geometry.get_original_shape() != opened_calib.geometry.get_original_shape() or (self.image_geometry.offset != opened_calib.geometry.offset and 'image_offset' not in self.image_assumptions):
                    osize = self.image_geometry.get_original_shape()
                    try:
                        opened_calib.set_detector_window((self.image_geometry.offset[0],self.image_geometry.offset[1],osize[0],osize[1]))
                    except NoSubviews:
                        raise UserWarning('The detector area according to the image metadata ({:d}x{:d} pixels at offset {:d}x{:d}) does not contain any image according to the calibration! Perhaps the image detector offset is set incorrectly?'.format(osize[0],osize[1],self.image_geometry.offset[0],self.image_geometry.offset[1]))

            if opened_calib.image is not None:
                self.movement_correction_button.setEnabled(True)
            else:
                self.movement_correction_button.setEnabled(False)

            if self.cadmodel is not None:
                self.im_mapping_checkbox.setEnabled(True)

        self.calibration = opened_calib
        self.image_geometry = self.calibration.geometry
        self.calib_name.setText(os.path.split(self.calibration.filename)[1].replace('.ccc',''))
        self.cal_props_button.setEnabled(True)

        self.overlay_checkbox.setEnabled(True)
        self.im_mapping_checkbox.setEnabled(True)
        self.reset_view_button.setEnabled(True)
        self.overlay = None
        self.im_projection = None

        if self.image is not None:
            if self.enhance_checkbox.isChecked():
                self.interactor2d.set_image(enhance_image(self.image_geometry.original_to_display_image(self.image)))
            elif self.cmap_checkbox.isChecked():
                self.update_cmap()
            else:
                self.interactor2d.set_image(self.image_geometry.original_to_display_image(self.image))

        self.interactor2d.set_subview_lookup(self.calibration.n_subviews,self.calibration.subview_lookup)

        if opened_calib.cad_config is not None:
            cconfig = opened_calib.cad_config
            if self.cadmodel is not None and self.cadmodel.machine_name == cconfig['model_name'] and self.cadmodel.model_variant == cconfig['model_variant']:
                keep_model = True
            else:
                keep_model = False

            if keep_model:
                try:
                    self.cadmodel.enable_only(cconfig['enabled_features'])
                except Exception as e:
                    if 'Unknown feature' not in str(e):
                        raise
                self.update_feature_tree_checks()
            else:
                cconfig = opened_calib.cad_config
                load_model = True
                try:
                    name_index = sorted(self.model_list.keys()).index(cconfig['model_name'])
                    self.model_name.setCurrentIndex(name_index)
                    variant_index = self.model_list[ cconfig['model_name'] ][1].index(cconfig['model_variant'])
                    self.model_variant.setCurrentIndex(variant_index)
                except ValueError:
                    self.model_name.setCurrentIndex(-1)
                    load_model=False


                if load_model:
                    try:
                        self.load_model(featurelist=cconfig['enabled_features'])
                    except Exception as e:
                        self.cadmodel = None
                        if 'Unknown feature' not in str(e):
                            raise

                # I'm not sure why I have to call this twice to get it to work properly.
                # On the first call it almost works, but not quite accurately. Then
                # on the second call onwards it's fine. Should be investigated further.
                self.set_view_from_calib(self.calibration,0)
                self.set_view_from_calib(self.calibration,0)

        if self.overlay_checkbox.isChecked():
            self.overlay_checkbox.setChecked(False)
            self.overlay_checkbox.setChecked(True)

        if self.cursor_ids['3d'] is not None:
            self.update_from_3d(self.coords_3d)
        else:
            self.coords_2d = [None] * self.calibration.n_subviews



    def unload_calib(self):

        if self.calibration is not None:
            self.calibration = None
            self.calib_name.setText('No Calibration Loaded.')
            self.overlay_checkbox.setChecked(False)
            self.overlay_checkbox.setEnabled(False)
            self.im_mapping_checkbox.setChecked(False)
            self.im_mapping_checkbox.setEnabled(False)
            self.reset_view_button.setEnabled(False)
            self.overlay = None
            self.im_projection = None
            self.cal_props_button.setEnabled(False)

            if self.cursor_ids['2d']['visible'] is not None:
                self.interactor2d.remove_active_cursor(self.cursor_ids['2d']['visible'])
                self.cursor_ids['2d']['visible'] = None

            if self.cursor_ids['2d']['hidden'] is not None:
                self.interactor2d.remove_active_cursor(self.cursor_ids['2d']['hidden'])
                self.cursor_ids['2d']['hidden'] = None


    def set_view_to_cursor(self):
        cursorpos = self.interactor3d.get_cursor_coords(self.cursor_ids['3d'])

        pupilpos = self.calibration.get_pupilpos(subview=0)
        campos_vect = (pupilpos - cursorpos)
        campos_vect = campos_vect / np.sqrt(np.sum(campos_vect**2))

        self.camera_3d.SetFocalPoint(cursorpos)
        self.camera_3d.SetViewAngle(60)

        self.camera_3d.SetPosition(cursorpos + 0.3*campos_vect)

        self.update_viewport_info()
        self.cam_roll.setValue(0)
        self.refresh_3d()


    def on_model_load(self):
        # Enable the other tabs!
        self.tabWidget.setTabEnabled(2,True)
        #self.tabWidget.setTabEnabled(2,True)
        #self.tabWidget.setTabEnabled(3,True)
        if self.image is not None:
            self.tabWidget.setTabEnabled(3,True)

        self.app.setOverrideCursor(qt.QCursor(qt.Qt.WaitCursor))
        self.statusbar.showMessage('Generating model octree...')
        self.cadmodel.get_cell_locator()
        self.statusbar.clearMessage()
        self.app.restoreOverrideCursor()
        self.overlay = None
        if self.image is not None and self.calibration is not None:
            self.im_mapping_checkbox.setEnabled(True)
        if self.overlay_checkbox.isChecked():
            self.overlay_checkbox.setChecked(False)
            self.overlay_checkbox.setChecked(True)



    def on_load_image(self,newim):

        image = scale_to_8bit(newim['image_data'])

        newim_geometry = CoordTransformer(offset=newim['image_offset'],paspect=newim['pixel_aspect'])
        newim_geometry.set_transform_actions(newim['transform_actions'])
        newim_geometry.set_image_shape(newim['image_data'].shape[1],newim['image_data'].shape[0],coords=newim['coords'])

        reset_overlay = False

        if self.calibration is not None:

            if newim['coords'].lower() == 'original':
                size_mismatch = newim_geometry.get_original_shape() != self.calibration.geometry.get_original_shape()
            else:
                size_mismatch = newim_geometry.get_display_shape() != self.calibration.geometry.get_display_shape()

            if 'image_offset' not in newim['assumptions']:
                size_mismatch = size_mismatch | (newim_geometry.offset[0] != self.calibration.geometry.offset[0]) | (newim_geometry.offset[1] != self.calibration.geometry.offset[1])

            if size_mismatch:

                if 'coords' in newim['assumptions']:

                    if newim_geometry.get_display_shape() == self.calibration.geometry.get_original_shape():
                        newim_geometry = self.calibration.geometry
                        newim['coords'] = 'original'

                new_geom = copy.deepcopy(self.calibration.geometry)
                new_geom.set_image_shape(image.shape[1],image.shape[0],coords=newim['coords'])
                new_geom.set_offset(*newim_geometry.offset)
                osize = new_geom.get_original_shape()

                self.calibration.geometry.set_pixel_aspect(new_geom.pixel_aspectratio,relative_to='original')
                try:
                    self.calibration.set_detector_window((newim_geometry.offset[0], newim_geometry.offset[1], osize[0], osize[1]))
                except NoSubviews:
                    raise UserWarning('The detector area according to the image metadata ({:d}x{:d} pixels at offset {:d}x{:d}) does not contain any image according to the calibration! Perhaps the image detector offset is set incorrectly?'.format(osize[0], osize[1], self.image_geometry.offset[0], self.image_geometry.offset[1]))


                self.overlay = None
                if self.overlay_checkbox.isChecked():
                    self.overlay_checkbox.setChecked(False)
                    reset_overlay = True

            self.image_geometry = self.calibration.geometry

            if self.calibration.image is not None:
                self.movement_correction_button.setEnabled(True)

            if self.cadmodel is not None:
                self.im_mapping_checkbox.setEnabled(True)

        else:
            self.image_geometry = newim_geometry

        if newim['coords'].lower() == 'display':
            self.image = self.image_geometry.display_to_original_image(image)
        else:
            self.image = image

        self.image_assumptions = newim['assumptions']
        self.image_load_coords = newim['coords']


        if self.calibration is not None:
            self.interactor2d.set_image(self.image_geometry.original_to_display_image(self.image),n_subviews=self.calibration.n_subviews,subview_lookup=self.calibration.subview_lookup)
        else:
            self.interactor2d.set_image(self.image_geometry.original_to_display_image(self.image))

        self.overlay_checkbox.setChecked(reset_overlay)
        self.image_settings.show()

        if self.enhance_checkbox.isChecked():
            self.enhance_checkbox.setChecked(False)
            self.enhance_checkbox.setChecked(True)

        try:
            coords3d = self.interactor3d.get_cursor_coords(self.cursor_ids['3d'])
            self.update_from_3d(coords3d)
        except KeyError as e:
            pass

        self.update_image_info_string(self.image,self.image_geometry)
        self.app.restoreOverrideCursor()
        self.statusbar.clearMessage()

        if self.cadmodel is not None:
            self.tabWidget.setTabEnabled(3,True)

        if self.overlay_checkbox.isChecked():
            self.overlay_checkbox.setChecked(False)
            self.overlay_checkbox.setChecked(True)

        if self.cmap_checkbox.isChecked():
            self.update_cmap()


    def update_image_info_string(self, im_array, geometry):

        if np.any(np.array(geometry.get_display_shape()) != np.array(geometry.get_original_shape())):
            info_str = '{0:d} x {1:d} pixels ({2:.1f} MP) [ As Displayed ]<br>{3:d} x {4:d} pixels ({5:.1f} MP) [ Raw Data ]<br>'.format(
                geometry.get_display_shape()[0], geometry.get_display_shape()[1],
                np.prod(geometry.get_display_shape()) / 1e6, geometry.get_original_shape()[0],
                geometry.get_original_shape()[1], np.prod(geometry.get_original_shape()) / 1e6)
        else:
            info_str = '{0:d} x {1:d} pixels ({2:.1f} MP)<br>'.format(geometry.get_display_shape()[0],
                                                                      geometry.get_display_shape()[1],
                                                                      np.prod(geometry.get_display_shape()) / 1e6)

        if len(im_array.shape) == 2:
            info_str = info_str + 'Monochrome'
            self.cmap_checkbox.setEnabled(True)
            self.cmap_min.blockSignals(True)
            self.cmap_max.blockSignals(True)
            self.cmap_min.setMinimum(im_array.min())
            self.cmap_min.setMaximum(im_array.max())
            self.cmap_min.setValue(im_array.min())
            self.cmap_max.setMinimum(im_array.min())
            self.cmap_max.setMaximum(im_array.max())
            self.cmap_max.setValue(im_array.max())
            self.cmap_min.blockSignals(False)
            self.cmap_max.blockSignals(False)
            self.cmap_checkbox.setToolTip('')
        elif len(im_array.shape) == 3:
            if np.all(im_array[:, :, 0] == im_array[:, :, 1]) and np.all(im_array[:, :, 2] == im_array[:, :, 1]):
                info_str = info_str + 'Monochrome'
                self.cmap_min.blockSignals(True)
                self.cmap_max.blockSignals(True)
                self.cmap_checkbox.setEnabled(True)
                self.cmap_checkbox.setEnabled(True)
                self.cmap_min.setMinimum(im_array.min())
                self.cmap_min.setMaximum(im_array.max())
                self.cmap_min.setValue(im_array.min())
                self.cmap_max.setMinimum(im_array.min())
                self.cmap_max.setMaximum(im_array.max())
                self.cmap_max.setValue(im_array.max())
                self.cmap_min.blockSignals(False)
                self.cmap_max.blockSignals(False)
                self.cmap_checkbox.setToolTip('')
            else:
                info_str = info_str + 'RGB Colour'
                self.cmap_checkbox.setChecked(False)
                self.cmap_checkbox.setEnabled(False)
                self.cmap_checkbox.setToolTip('Colour mapping can only be applied to monochrome images.')
            if im_array.shape[2] == 4:
                info_str = info_str + ' + Transparency'

        self.image_info.setText(info_str)

    def update_cmap(self,data=None):

        if self.sender() is self.cmap_checkbox:
            if self.cmap_checkbox.isChecked():
                self.enhance_checkbox.setChecked(False)
                self.cmap_options.show()
            else:
                self.cmap_options.hide()
                self.toggle_enhancement()
                return

        self.cmap_min.blockSignals(True)
        self.cmap_max.blockSignals(True)
        self.cmap_min.setMaximum(self.cmap_max.value() - 1)
        self.cmap_max.setMinimum(self.cmap_min.value() + 1)
        self.cmap_min.blockSignals(False)
        self.cmap_max.blockSignals(False)

        image = self.image.astype(np.float32)
        if len(image.shape) > 2:
            image = image[:,:,0]

        cmin = self.cmap_min.value()
        cmax = self.cmap_max.value()

        cmap = cmaps[self.cmap_dropdown.currentIndex()][1]

        im = (image - cmin) / (cmax - cmin)
        im[im > 1] = 1
        im[im < 0] = 0
        cmapped_im = cmap(im)[:,:,:3]
        cmapped_im = (255 * cmapped_im).astype(np.uint8)

        if self.calibration is not None:
            transformer = self.calibration.geometry
        else:
            transformer = self.image_geometry

        self.interactor2d.set_image(transformer.original_to_display_image(cmapped_im), hold_position=True)
        if self.calibration is not None:
            self.interactor2d.set_subview_lookup(self.calibration.n_subviews,self.calibration.subview_lookup)

        if self.overlay_checkbox.isChecked():
            self.overlay_checkbox.setChecked(False)
            self.overlay_checkbox.setChecked(True)

        self.update_mapped_image()


    def update_overlay(self,data=None):

        # Clear the existing overlay image to force it to re-render if the user has changed between solid / wireframe
        if self.sender() is self.overlay_type_box:
            self.overlay = None

        # Show or hide extra controls depending on what overlays are enabled
        self.overlay_settings.setVisible(self.overlay_checkbox.isChecked())

        if self.overlay_checkbox.isChecked():

            if self.overlay is None:
                self.overlay = self.render_overlay_image(self.overlay_type_box.currentIndex() == 0)

            # Apply desired colour
            im = (self.overlay * np.tile(np.array(self.overlay_colour)[np.newaxis, np.newaxis, :], self.overlay.shape[:2] + (1,))).astype(np.uint8)

            if self.mov_correction is not None:
                self.interactor2d.set_overlay_image(self.mov_correction.warp_ref_to_moved(im)[0])
            else:
                self.interactor2d.set_overlay_image(im)


        else:
            self.interactor2d.set_overlay_image(None)

        self.refresh_2d()


    def change_overlay_colour(self):

        if self.sender() is self.overlay_opacity_slider:
            new_colour = self.overlay_colour[:3]

        else:
            old_colour = self.overlay_colour
            new_colour = self.pick_colour(init_colour=old_colour)

        if new_colour is not None:

            new_colour = new_colour + [self.overlay_opacity_slider.value() / 100]
            self.overlay_colour = new_colour

            self.update_overlay()


    def render_overlay_image(self,wireframe):

        oversampling = int(np.ceil(min(1000/np.array(self.calibration.geometry.get_display_shape()))))

        self.statusbar.showMessage('Rendering CAD image overlay...')
        self.app.setOverrideCursor(qt.QCursor(qt.Qt.WaitCursor))
        self.app.processEvents()

        orig_colours = self.cadmodel.get_colour()

        self.cadmodel.set_wireframe(wireframe)
        self.cadmodel.set_colour((1, 1, 1))
        image = render_cam_view(self.cadmodel, self.calibration, transparency=True, verbose=False, aa=2,oversampling=oversampling)
        self.cadmodel.set_colour(orig_colours)
        self.cadmodel.set_wireframe(False)

        self.statusbar.clearMessage()
        self.app.restoreOverrideCursor()

        return image



    def toggle_enhancement(self):

        # Enable / disable adaptive histogram equalisation
        if self.enhance_checkbox.isChecked():
            self.cmap_checkbox.setChecked(False)
            image = enhance_image(self.image)
        else:
            image = self.image

        if self.calibration is not None:
            transformer = self.calibration.geometry
        else:
            transformer = self.image_geometry

        self.interactor2d.set_image(transformer.original_to_display_image(image),hold_position=True)
        if self.calibration is not None:
            self.interactor2d.set_subview_lookup(self.calibration.n_subviews,self.calibration.subview_lookup)

        self.update_mapped_image()


    def on_change_cad_features(self):
        self.app.setOverrideCursor(qt.QCursor(qt.Qt.WaitCursor))
        self.statusbar.showMessage('Updating model octree...')
        self.cadmodel.get_cell_locator()
        self.statusbar.clearMessage()
        self.app.restoreOverrideCursor()
        if self.overlay_checkbox.isChecked():
            self.overlay_checkbox.setChecked(False)
        self.overlay = None
        self.im_mapping_checkbox.setChecked(False)
        self.im_projection = None


    def update_position_info(self,coords_2d,coords_3d,visible):

        model_extent = self.cadmodel.get_extent()
        model_size = model_extent[1::2] - model_extent[::2]
        max_ray_length = model_size.max() * 4

        cadinfo_str = self.cadmodel.format_coord(coords_3d)

        sightline_info_string = ''

        iminfo_str = ''

        sightline_fieldnames_str = ''

        impos_fieldnames_str = ''

        self.cursor_closeup_button.setEnabled(False)

        for field_index in range(self.calibration.n_subviews):

            if field_index > 0:
                impos_fieldnames_str = impos_fieldnames_str + '<br><br>'
                sightline_fieldnames_str = sightline_fieldnames_str + '<br><br>'
                iminfo_str = iminfo_str + '<br><br>'
                sightline_info_string = sightline_info_string + '<br><br>'

            sightline_exists = False


            impos_fieldnames_str = impos_fieldnames_str + '[{:s}]&nbsp;'.format( self.calibration.subview_names[field_index] )

            if np.any(np.isnan(coords_2d[field_index][0])):
                iminfo_str = iminfo_str + 'Cursor outside field of view.'
            elif not visible[field_index]:
                iminfo_str = iminfo_str +  'Cursor hidden from view.'
            else:
                iminfo_str = iminfo_str + 'X,Y : ( {:.0f} , {:.0f} ) px'.format(coords_2d[field_index][0][0],coords_2d[field_index][0][1])
                sightline_exists = True
                self.cursor_closeup_button.setEnabled(True)


            sightline_fieldnames_str = sightline_fieldnames_str + '[{:s}]&nbsp;'.format( self.calibration.subview_names[field_index] )
            if sightline_exists:
                sightline_fieldnames_str = sightline_fieldnames_str + '<br><br>'
                pupilpos = self.calibration.get_pupilpos(subview=field_index)
                sightline = coords_3d - pupilpos
                sdir = sightline / np.sqrt(np.sum(sightline**2))


                sightline_info_string = sightline_info_string + 'Origin X,Y,Z : ( {:.3f} , {:.3f} , {:.3f} )<br>'.format(pupilpos[0],pupilpos[1],pupilpos[2])
                sightline_info_string = sightline_info_string + 'Direction X,Y,Z : ( {:.3f} , {:.3f} , {:.3f} )<br>'.format(sdir[0],sdir[1],sdir[2])
                if np.sqrt(np.sum(sightline**2)) < (max_ray_length-1e-3):
                    sightline_info_string = sightline_info_string  +'Distance from camera to cursor: {:.3f} m'.format(np.sqrt(np.sum(sightline**2)))
                else:
                    sightline_info_string = sightline_info_string + 'Sight line does not inersect CAD model.'
                    cadinfo_str = 'Sight line does not intersect CAD model.'
            else:

                sightline_info_string = sightline_info_string + 'No line-of-sight to cursor'

        if self.calibration.n_subviews > 1:
            self.impos_fieldnames.setText(impos_fieldnames_str)
            self.impos_fieldnames.show()
            self.sightline_fieldnames.setText(sightline_fieldnames_str)
            self.sightline_fieldnames.show()
        else:
            self.impos_fieldnames.hide()
            self.sightline_fieldnames.hide()

        self.impos_info.setText(iminfo_str)
        self.sightline_info.setText(sightline_info_string)
        self.cadpos_info.setText(cadinfo_str)


'''
* Copyright 2015-2021 European Atomic Energy Community (EURATOM)
*
* Licensed under the EUPL, Version 1.1 or - as soon they
  will be approved by the European Commission - subsequent
  versions of the EUPL (the "Licence");
* You may not use this work except in compliance with the
  Licence.
* You may obtain a copy of the Licence at:
*
* https://joinup.ec.europa.eu/software/page/eupl
*
* Unless required by applicable law or agreed to in
  writing, software distributed under the Licence is
  distributed on an "AS IS" basis,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
  express or implied.
* See the Licence for the specific language governing
  permissions and limitations under the Licence.
'''

import zipfile
import tempfile
import os
import shutil
import hashlib
import atexit

from .misc import import_source,unload_source

# Get a list of directory contents, including sub-directories.
# Returned list has absolute paths.
def listdir(path):

    filelist = []

    fnames = sorted(os.listdir(path))

    for fname in [ os.path.join(path,f) for f in fnames]:

        # Recursively do sub-directories
        if os.path.isdir(fname):
            filelist = filelist + listdir(fname)
        else:
            filelist.append(fname)

    return filelist


def md5_file(filename):

    blocksize = 65536
    hasher = hashlib.md5()
    with open(filename, 'rb') as f:
        buf = f.read(blocksize)
        while len(buf) > 0:
            hasher.update(buf)
            buf = f.read(blocksize)
    return hasher.digest()


# Class for Zip file based save files.
class ZipSaveFile():

    def __init__(self,fname,mode='r',ignore_pyc=True):

        if 'w' in mode and 's' in mode:
            raise ValueError('Invalid mode "{:s}"'.format(mode))

        self.filename = os.path.abspath(fname)
        self.ignore_pyc = ignore_pyc
        self.mode = mode
        self.pypaths = []
        self.is_open = False
        self.open(self.mode)
        self.usermodule = None

        atexit.register(self.close)


    def open(self,mode):

        # Maybe we're re-opening the file, in which case
        # we need to close first.
        if self.is_open:
            self.close()

        self.mode = mode

        # Check the file exists, and that if we're going to try to write to it,
        # that we have the necessary permissions.
        if 'w' in self.mode:
            if os.path.isfile(self.filename):
                if not os.access(self.filename,os.W_OK):
                    raise IOError('Write permission denied on {:s}'.format(self.filename))
            elif os.path.isdir(self.filename):
                raise IOError('Specified filename already exists but is a directory!')
            else:
                savepath = os.path.split(self.filename)[0]
                if not os.access(savepath,os.W_OK):
                    raise IOError('Cannot write to directory {:s}'.format(savepath))

        if 'r' in self.mode and 'w' not in self.mode:
            if not os.path.isfile(self.filename):
                raise IOError('No such file: {:s}'.format(self.filename))


        # Create a temporary directory which we'll use 
        # to extract our ZIP while we work with its contents.
        self.tempdir = tempfile.mkdtemp()

        if 'r' in self.mode:

            try:
                with zipfile.ZipFile(self.filename,'r') as zf:

                    if 's' not in mode:
                        loadlist = zf.namelist()    
                    else:
                        loadlist = [name for name in zf.namelist() if not name.startswith('.large/')]

                    zf.extractall(self.tempdir,members=loadlist)
            
            except:
                if 'w' not in self.mode:
                    shutil.rmtree(self.tempdir)
                    raise

        self.file_handles = []
        self.is_open = True
        self.initial_hashes = self.get_hashes()


    def is_readonly(self):

        return not os.access(self.filename,os.W_OK)


    def get_hashes(self):

        if self.is_open:
            hashes = []
            for fname in self.list_contents():
                hashes.append( (fname,md5_file(os.path.join(self.tempdir,fname))) )

            return hashes
        else:
            raise Exception('File is not open!')


    def close(self,discard_changes=False):
        
        if self.is_open:
            for h in self.file_handles:
                h.close()

            # If we're in write mode, and the file contents have been modified since being loaded,
            # we need to re-save the ZIP file with the new contents.
            if 'w' in self.mode and not discard_changes and self.get_hashes() != self.initial_hashes:
                self.update()

            # Make sure we properly unload any user code
            try:
                unload_source(os.path.join(self.tempdir,'usercode'))
            except FileNotFoundError:
                pass

            self.is_open = False
            
            # Tidy up the temp directory after ourselves
            shutil.rmtree(self.tempdir)
            self.tempdir = None

        atexit.unregister(self.close)



    def update(self):

        with zipfile.ZipFile(self.filename,'w',zipfile.ZIP_DEFLATED,True) as zf:
            for fname in listdir(self.tempdir):
                zf.write(fname,os.path.relpath(fname,self.tempdir))


    # Open a file inside the zip for doing stuff with.
    # Reurns a file handle to the opened file.
    def open_file(self,fname,mode):

        if not self.is_open:
            self.open()

        if 'r' in mode and fname not in self.list_contents():
            raise IOError('File "{:s}" not in here!'.format(fname))

        if 'w' in mode and 'w' not in self.mode:
            raise IOError('File is open in read only mode!')

        h = open( os.path.join(self.tempdir,fname) , mode )

        self.file_handles.append(h)

        return h


    # If the is Python code in the zip either in usercode.py or a 
    # directory called usercode/, return it as a python module.
    def get_usercode(self):

        if not self.is_open:
            self.open()

        if 'r' not in self.mode or not self.is_open:
            raise IOError('File not open in read mode!')

        if os.path.join('usercode','__init__.py') in self.list_contents():
            return import_source(os.path.join(self.tempdir,'usercode'))
        elif 'usercode.py' in self.list_contents():
            return import_source(os.path.join(self.tempdir, 'usercode.py'))
        else:
            return None




    # Get a list of the files within.
    def list_contents(self):

        if not self.is_open:
            self.open()
        
        if self.ignore_pyc:
            return [os.path.relpath(fname,self.tempdir) for fname in listdir(self.tempdir) if not fname.endswith('.pyc')]
        else:
            return [os.path.relpath(fname,self.tempdir) for fname in listdir(self.tempdir)]


    # Add a file or directory to the archive.
    def add(self,from_path,to_path=None,replace=False):

        if not self.is_open:
            self.open()

        if 'w' not in self.mode:
            raise IOError('File is open in read-only mode!')

        if not (os.path.isdir(from_path) or os.path.isfile(from_path)):
            raise IOError('No such file or directory "{:s}"'.format(from_path))

        if to_path is None:
            dst_path = os.path.join(self.tempdir, from_path.split(os.sep)[-1] )
        else:
            dst_path = os.path.join(self.tempdir, to_path )

        if os.path.isdir(dst_path):
            if replace:
                shutil.rmtree(dist_path)
            else:
                raise IOError('This path already exists in this file! Use replace=True to allow overwriting.')
        elif os.path.isfile(dst_path):
            if replace:
                os.remove(dst_path)
            else:
                raise IOError('This file already exists in this file! Use replace=True to allow overwriting.')
        
        dst_folder = os.path.split(dst_path)[0]
        if not os.path.isdir(dst_folder):
            os.makedirs(dst_folder)

        if os.path.isfile(from_path):
            shutil.copy2(from_path,dst_path)
        elif os.path.isdir(from_path):
            shutil.copytree(from_path,dst_path)


    # Add user python code to the arvhive.
    def add_usercode(self,usercode_path,replace = False):

        if not self.is_open:
            self.open()

        if 'w' not in self.mode:
            raise IOError('File is open in read-only mode!')

        if self.get_usercode() and not replace:
                raise IOError('File already contains user code! Use replace=True to allow overwriting.')

        self.clear_usercode()
        
        if os.path.isfile(usercode_path):
            self.add(usercode_path,'usercode.py')
        else:
            self.add(usercode_path,'usercode')



    # Add user python code to the arvhive.
    def clear_usercode(self):

        if not self.is_open:
            self.open()

        if 'w' not in self.mode:
            raise IOError('File is open in read-only mode!')

        unload_source(os.path.join(self.tempdir,'usercode'))

        try:
            self.remove('usercode')
        except IOError:
            pass
        try:
            self.remove('usercode.py')
        except IOError:
            pass        


    # Remove a named file or folder from the archive
    def remove(self,fname):

        if 'w' not in self.mode:
            raise IOError('File is open in read-only mode!')

        fullpath = os.path.join(self.tempdir, fname)

        if os.path.isfile( fullpath ):
            os.remove(fullpath)
        elif os.path.isdir( fullpath ):
            shutil.rmtree( fullpath )
        else:
            raise IOError('File or directory "{:s}" not in here!'.format(fname))

    # Return the temporary path for manually playing with / using contents.
    def get_temp_path(self):

        return self.tempdir


    def mkdir(self,dirname):

        if not os.path.exists(dirname):
            os.makedirs(os.path.join(self.tempdir,dirname))


    # For context management
    def __enter__(self):

        if not self.is_open:
            self.open()

        return self


    # For context management
    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
        if exc_type is not None:
            raise


    # If we forget to close the object properly
    # before it's garbage collected or whatever, make sure the 
    # temp files get cleaned up.
    def __del__(self):
        self.close()


'''
* Copyright 2015-2020 European Atomic Energy Community (EURATOM)
*
* Licensed under the EUPL, Version 1.1 or - as soon they
  will be approved by the European Commission - subsequent
  versions of the EUPL (the "Licence");
* You may not use this work except in compliance with the
  Licence.
* You may obtain a copy of the Licence at:
*
* https://joinup.ec.europa.eu/software/page/eupl
*
* Unless required by applicable law or agreed to in
  writing, software distributed under the Licence is
  distributed on an "AS IS" basis,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
  express or implied.
* See the Licence for the specific language governing
  permissions and limitations under the Licence.
'''

import cv2
import numpy as np
import copy


class CoordTransformer:
    """
    Class to handle coordinate transformations between 'display' and 'original' image coordinates.
    """
    
    def __init__(self,transform_actions=[],orig_x=None,orig_y=None,paspect=1.,offset=(0,0)):

        self.set_transform_actions(transform_actions)
        self.x_pixels = orig_x
        self.y_pixels = orig_y
        self.offset = offset
        self.pixel_aspectratio = paspect



    def set_transform_actions(self,transform_actions):
        """
        Set the actions to perform on an image to change it from original to display coordinates.

        Parameters:

            transform_actions (list of str) : List of strings specifying image transform actions, in the order \\
                                              they should be performed. Allowed strings are: 'flip_up_down', \\
                                              'flip_left_right','rotate_clockwise_90','rotate_clockwise_180', \\
                                              or 'rotate_clockwise_270'
        """
        self.transform_actions = []
        for action in transform_actions:
            if action.lower() in ['flip_up_down','flip_left_right','rotate_clockwise_90','rotate_clockwise_180','rotate_clockwise_270']:
                self.transform_actions.append(action)
            elif action in ['','rotate_clockwise_0']:
                pass            
            else:
                raise Exception('Unknown transformation action "' + action + '"')
                


    def get_transform_actions(self):
        """
        Returns a list of strings specifying the actions to perform on an image to change it from original to display coordinates,
        in the order they are performed.
        """
        return copy.copy(self.transform_actions)

    def get_image_shape(self,coords):
        """
        A convenience function to return the image shape depending on a string input for the coords.

        Params:
            coords (str) : 'Original' or 'Display' , whether to return the original or display coords

        Returns:
            Tuple with (width, height) image shape in pixels
        """
        if coords.lower() == 'original':
            return self.get_original_shape()
        elif coords.lower() == 'display':
            return self.get_display_shape()


    def set_image_shape(self,w,h,coords='Original'):
        """
        Set the shape of the image to be transformed.

        Parameters:
            w (int)      : Image witdh in pixels
            h (int)      : Image height in pixels
            coords (str) : 'Original' or 'Display', whether the specified with and height are the display or original shape
        """
        if coords.lower() == 'display':

            shape = list(self.display_to_original_shape([w,h]))

            if self._is_sideways():
                self.pixel_aspectratio = w/shape[1]
            else:
                self.pixel_aspectratio = h/shape[1]

        else:
            shape = [w, h]
            self.pixel_aspectratio = np.round(h * self.pixel_aspectratio) / shape[1]

        self.x_pixels = int(shape[0])
        self.y_pixels = int(shape[1])



    def _is_sideways(self):
        """
        Returns true if the image is rotated by 90 or 270 degrees compared to the
        raw sensor image, otherwise returns false
        """
        sideways = False
        for action in self.transform_actions:
            if action.lower() in ['rotate_clockwise_90', 'rotate_clockwise_270']:
                sideways = not sideways

        return sideways


    def set_offset(self,x_offset,y_offset):
        """
        Specify the offset of the image top-left from the detector top-left (for CMOS sensors with smaller
        redout windows.

        Parameters:

            x_offset (int) : Horizontal offset in pixels
            y_offset (int) : Vertical offset in pixels
        """
        self.offset = (x_offset,y_offset)



    def add_transform_action(self,transform_action):
        """
        Specify an additional action to change the image from original to display coordinates.
        The added action will be performed last.

        Parameters:

            transform_action (str) : String specifying the transform action (for allowed values see `set_transform_actions()`
        """
        transform_action = transform_action.lower()

        if transform_action not in ['flip_up_down','flip_left_right','rotate_clockwise_90','rotate_clockwise_180','rotate_clockwise_270']:
            raise ValueError('Unknown transform action {:s}'.format(transform_action))

        if len(self.transform_actions) == 0:
            self.transform_actions = [transform_action]
        else:
            if transform_action == 'flip_up_down' and self.transform_actions[-1] == 'flip_up_down':
                del self.transform_actions[-1]
            elif transform_action == 'flip_left_right' and self.transform_actions[-1] == 'flip_left_right':
                del self.transform_actions[-1]
            elif 'rotate_clockwise' in transform_action and 'rotate_clockwise' in self.transform_actions[-1]:
                current_angle = int(self.transform_actions[-1].split('_')[2])
                del self.transform_actions[-1]
                new_angle = int(transform_action.split('_')[2])
                total_angle = current_angle + new_angle
                if total_angle > 270:
                    total_angle = total_angle - 360
                if new_angle > 0:
                    self.transform_actions.append('rotate_clockwise_{:d}'.format(total_angle))
            else:
                self.transform_actions.append(transform_action)




    def set_pixel_aspect(self,pixel_aspect,relative_to='display',absolute=True):
        """
        Set the pixel aspect ratio of the camera (pixel height / pixel width) for cameras with non-square pixels.

        Parameters:

            pixel_aspect (float) : Ratio of pixel height / pixel width.

            relative_to (str)    : 'Display' or 'Original', whether the height/width is specified with the image \
                                   in its original or display coordinates

            absolute (bool)      : Whether the specified ratio is specified in absolute terms (default) \
                                   or relative to the currently set pixel aspect ratio (i.e. only has an effect \
                                   if the aspect ratio is already set != 1).
        """
        if absolute:
            ref_aspect = 1.
        else:
            ref_aspect = float(self.pixel_aspectratio)

        if relative_to.lower() == 'original':
            self.pixel_aspectratio = pixel_aspect * ref_aspect
        else:
            
            if self._is_sideways():
                self.pixel_aspectratio = ref_aspect/pixel_aspect
            else:
                self.pixel_aspectratio = ref_aspect*pixel_aspect



    def original_to_display_shape(self,shape):
        """
        Based on the transform actions and pixel aspect ratio, get the display image shape for a
        given original image shape.

        Parameters:

           shape (sequence) : 2-element sequence specifying the original image (width,height)

        Returns:

           Tuple : 2-element tuple specifying the displayed image (width,height).
        """
        shape = np.array(shape,dtype=np.float32)
        shape[1] = shape[1] * self.pixel_aspectratio

        if self._is_sideways():
            shape = shape[::-1]

        return tuple(np.round(shape).astype(int))



    def display_to_original_shape(self,shape):
        """
        Based on the transform actions and pixel aspect ratio, get the original image shape for a
        given displayed image shape.

        Parameters:

           shape (sequence) : 2-element sequence specifying the displayed image (width,height)

        Returns:

           Tuple : 2-element tuple specifying the original image (width,height).
        """
        shape = np.array(shape,dtype=np.float32)

        if self._is_sideways():
            shape = shape[::-1]

        shape[1] = shape[1] / self.pixel_aspectratio

        return tuple(np.round(shape).astype(int))




    def original_to_display_image(self,image,interpolation='nearest'):
        """
        Transform an image from original to display orientation.

        Parameters:

            image (np.ndarray)  : Array containing the image in original orientation.

            interpolation (str) : Interpolation method, allowed strings are 'nearest' or 'cubic'.


        Returns:

            np.ndarray : Array containing the image in display orientation.
        """
        if interpolation.lower() == 'nearest':
            interp_method = cv2.INTER_NEAREST
        elif interpolation.lower() == 'cubic':
            interp_method = cv2.INTER_CUBIC
        else:
            raise ValueError('Interpolation method must be "nearest" or "cubic".')

        expected_size = np.array(self.get_original_shape())
        im_size = np.array(image.shape[1::-1])
        ratio = expected_size / im_size

        if np.abs(ratio[0]-ratio[1]) < 1e-5:
            binning = ratio[0]
        else:
            raise Exception('Expected (multiple of) {:d}x{:d} pixel image, got {:d}x{:d}!'.format(expected_size[0],expected_size[1],image.shape[1],image.shape[0]))

        data_out = image.copy()


        for action in self.transform_actions:
            if action.lower() == 'flip_up_down':
                data_out = np.flipud(data_out)
            elif action.lower() == 'flip_left_right':
                data_out = np.fliplr(data_out)
            elif action.lower() == 'rotate_clockwise_90':
                data_out = np.rot90(data_out,k=3)
            elif action.lower() == 'rotate_clockwise_180':
                data_out = np.rot90(data_out,k=2)
            elif action.lower() == 'rotate_clockwise_270':
                data_out = np.rot90(data_out,k=1)

        out_shape = self.get_display_shape()

        if data_out.shape[0] != int(out_shape[1]/binning) or data_out.shape[1] != int(out_shape[0]/binning):
            data_out = cv2.resize(data_out,(int(out_shape[0]/binning),int(out_shape[1]/binning)),interpolation=interp_method)

        return data_out



    def display_to_original_image(self,image,interpolation='nearest'):
        """
        Transform an image from display to original orientation.

        Parameters:

            image (np.ndarray)  : Array containing the image in display orientation.

            interpolation (str) : Interpolation method, allowed strings are 'nearest' or 'cubic'.


        Returns:

            np.ndarray : Array containing the image in original orientation.
        """
        if interpolation.lower() == 'nearest':
            interp_method = cv2.INTER_NEAREST
        elif interpolation.lower() == 'cubic':
            interp_method = cv2.INTER_CUBIC
        else:
            raise ValueError('Interpolation method must be "nearest" or "cubic".')
            

        expected_size = np.array(self.get_display_shape())
        im_size = np.array(image.shape[1::-1])
        ratio = expected_size / im_size

        if np.abs(ratio[0]-ratio[1]) < 1e-5:
            binning = ratio[0]
        else:
            raise Exception('Expected (multiple of) {:d}x{:d} pixel image, got {:d}x{:d}!'.format(expected_size[0],expected_size[1],image.shape[1],image.shape[0]))

        data_out = image.copy()

        for action in reversed(self.transform_actions):
            if action.lower() == 'flip_up_down':
                data_out = np.flipud(data_out)
            elif action.lower() == 'flip_left_right':
                data_out = np.fliplr(data_out)
            elif action.lower() == 'rotate_clockwise_90':
                data_out = np.rot90(data_out,k=1)
            elif action.lower() == 'rotate_clockwise_180':
                data_out = np.rot90(data_out,k=2)
            elif action.lower() == 'rotate_clockwise_270':
                data_out = np.rot90(data_out,k=3)

        out_shape = self.get_original_shape()

        if data_out.shape[0] != int(out_shape[1]/binning) or data_out.shape[1] != int(out_shape[0]/binning):
            data_out = cv2.resize(data_out,(int(out_shape[0]/binning),int(out_shape[1]/binning)),interpolation=interp_method)

        return data_out



    def original_to_display_coords(self,x,y):
        """
        Given pixel coordinates in original coordinates, translate these to display coordinates.

        Parameters:

            x (float or np.ndarray) : Original x pixel coordinates. Can be a single value or array. \\
                                      x and y must be the same shape.

            y (float or np.ndarray) : Original y pixel coordinates. Can be a single value or array. \\
                                      x and y must be the same shape.


        Returns:

            x : Display x coordinates. Single value or array with the same shape as input x.

            y : Display y coordinates. Single value or array with the same shape as input y.
        """

        # Let's not overwrite the input arrays, just in case
        x_out = np.array(x)
        y_out = np.array(y) * self.pixel_aspectratio

        current_pixels = [self.x_pixels,int(self.y_pixels*self.pixel_aspectratio)]

        for action in self.transform_actions:
            if action.lower() == 'flip_up_down':
                y_out = (current_pixels[1]-1) - y_out
            elif action.lower() == 'flip_left_right':
                x_out = (current_pixels[0]-1) - x_out
            elif action.lower() == 'rotate_clockwise_90':
                # Temporary values...
                yt = y_out.copy()
                y_out = x_out.copy()

                x_out = (current_pixels[1]-1) - yt
                current_pixels = list(reversed(current_pixels))

            elif action.lower() == 'rotate_clockwise_180':

                y_out = (current_pixels[1]-1) - y_out
                x_out = (current_pixels[0]-1) - x_out
            elif action.lower() == 'rotate_clockwise_270':
                # Temporary values...
                yt = y_out.copy()

                y_out = (current_pixels[0]-1) - x_out
                x_out = yt
                current_pixels = list(reversed(current_pixels))

        if len(x_out.shape) == 0:
            x_out = float(x_out)
            y_out = float(y_out)

        return x_out,y_out



    def display_to_original_coords(self,x,y):
        """
        Given pixel coordinates in display coordinates, translate these to original coordinates.

        Parameters:

            x (float or np.ndarray) : Display x pixel coordinates. Can be a single value or array. \\
                                      x and y must be the same shape.

            y (float or np.ndarray) : Display y pixel coordinates. Can be a single value or array. \\
                                      x and y must be the same shape.


        Returns:

            x : Original x coordinates. Single value or array with the same shape as input x.

            y : Original y coordinates. Single value or array with the same shape as input y.
        """

        # Let's not overwrite the input arrays, just in case
        x_out = np.array(x)
        y_out = np.array(y)

        current_pixels = self.get_display_shape()

        for action in reversed(self.transform_actions):
            if action.lower() == 'flip_up_down':
                y_out = (current_pixels[1]-1) - y_out
            elif action.lower() == 'flip_left_right':
                x_out = (current_pixels[0]-1) - x_out
            elif action.lower() == 'rotate_clockwise_90':
                # Temporary values...
                yt = y_out.copy()

                y_out = (current_pixels[0]-1) - x_out
                x_out = yt
                current_pixels = list(reversed(current_pixels))
            elif action.lower() == 'rotate_clockwise_180':
                # Temporary values...
                y_out = (current_pixels[1]-1) - y_out
                x_out = (current_pixels[0]-1) - x_out
            elif action.lower() == 'rotate_clockwise_270':
                # Temporary values...
                yt = y_out.copy()
                y_out = x_out.copy()

                x_out = (current_pixels[1]-1) - yt
                current_pixels = list(reversed(current_pixels))

        y_out = y_out / self.pixel_aspectratio

        if len(x_out.shape) == 0:
            x_out = float(x_out)
            y_out = float(y_out)

        return x_out,y_out




    def get_display_shape(self):
        """
        Get the shape of the image in display orientation.

        Returns:

            Tuple specifying the (width,height) of the image in display orientation
        """

        display_shape = [self.x_pixels,int(np.round(self.y_pixels*self.pixel_aspectratio))]

        for action in self.transform_actions:
            if action.lower() in ['rotate_clockwise_90','rotate_clockwise_270']:
                display_shape.reverse()

        return tuple(display_shape)



    def get_original_shape(self):
        """
        Get the shape of the image in original orientation (as it comes from the camera).

        Returns:

            Tuple specifying the (width,height) of the image in original orientation
        """
        return (self.x_pixels,self.y_pixels)



    def display_to_original_pointpairs(self,pointpairs):
        """
        Given a set of point pairs in display coordinates, transform them to original coordinates.

        Parameters:

            pointpairs (calcam.PointPairs) : Point pairs object with the point pairs in display coordinates.

        Returns:

            calcam.PointPairs object with the point pairs in original coordinates.
        """
        if pointpairs is None:
            return None

        pp_out = copy.copy(pointpairs)

        for ipoint in range(pp_out.get_n_pointpairs()):
            for iview in range(pp_out.n_subviews):
                if pp_out.image_points[ipoint][iview] is not None:
                    pp_out.image_points[ipoint][iview] = tuple(self.display_to_original_coords(*pp_out.image_points[ipoint][iview]))

        return pp_out



    def original_to_display_pointpairs(self,pointpairs):
        """
        Given a set of point pairs in original coordinates, transform them to display coordinates.

        Parameters:

            pointpairs (calcam.PointPairs) : Point pairs object with the point pairs in original coordinates.

        Returns:

            calcam.PointPairs object with the point pairs in display coordinates.
        """
        if pointpairs is None:
            return None

        pp_out = copy.copy(pointpairs)

        for ipoint in range(pp_out.get_n_pointpairs()):
            for iview in range(pp_out.n_subviews):
                if pp_out.image_points[ipoint][iview] is not None:
                    pp_out.image_points[ipoint][iview] = tuple(self.original_to_display_coords(*pp_out.image_points[ipoint][iview]))

        return pp_out


'''
* Copyright 2015-2018 European Atomic Energy Community (EURATOM)
*
* Licensed under the EUPL, Version 1.1 or - as soon they
  will be approved by the European Commission - subsequent
  versions of the EUPL (the "Licence");
* You may not use this work except in compliance with the
  Licence.
* You may obtain a copy of the Licence at:
*
* https://joinup.ec.europa.eu/software/page/eupl
*
* Unless required by applicable law or agreed to in
  writing, software distributed under the Licence is
  distributed on an "AS IS" basis,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
  express or implied.
* See the Licence for the specific language governing
  permissions and limitations under the Licence.
'''

import cv2
import os

# A function which actually gets the image.
# This can take as many arguments as you want.
# It must return a dictionary, see below
def get_image(filename,coords,offset_x,offset_y):

    # Get the image data from a file
    dat = cv2.imread(filename)
    if dat is None:
        raise UserWarning('Could not read specified image file "' + filename + '"')

    # If it's a colour image, fix the colour channel order (openCV loads BGRA, for some reason)
    if len(dat.shape) == 3:

        # If we have a colour image swap the G and B channels, since OpenCV loads them in the wrong order.
        if dat.shape[2] == 3:
          
            dat[:,:,:3] = dat[:,:,2::-1]

            # If R, G and B channels are all the same, just return a monochrome image
            if (dat[:,:,0] == dat[:,:,1]).all() and (dat[:,:,0] == dat[:,:,2]).all():
                dat = dat[:,:,0] 


    return_dict = {
                    'image_data': dat,
                    'source': 'Loaded from image file {:s}'.format(os.path.split(filename)[-1]),
                    'coords':coords,
                    'image_offset':(offset_x,offset_y)
                    }

    return return_dict


display_name = 'Image File'
get_image_function = get_image
get_image_arguments =  [
                        {
                            'arg_name': 'filename',
                            'gui_label': 'File Name',
                            'type': 'filename',
                            'filter': 'Image Files (*.png *.jpg *.jpeg *.bmp *.jp2 *.tiff *.tif)'
                        },
                        {
                            'arg_name': 'coords',
                            'gui_label': 'Image Orientation',
                            'type': 'choice',
                            'choices': ['Display','Original'],
                            'default': 'Display'
                        },
                        {
                            'arg_name': 'offset_x',
                            'gui_label': 'Detector X Offset',
                            'type': 'int',
                            'limits':[0,1e4]
                        },
                        {
                            'arg_name': 'offset_y',
                            'gui_label': 'Detector Y Offset',
                            'type': 'int',
                            'limits': [0, 1e4]
                        },
                        ]


'''
* Copyright 2015-2019 European Atomic Energy Community (EURATOM)
*
* Licensed under the EUPL, Version 1.1 or - as soon they
will be approved by the European Commission - subsequent
versions of the EUPL (the "Licence");
* You may not use this work except in compliance with the
Licence.
* You may obtain a copy of the Licence at:
*
* https://joinup.ec.europa.eu/software/page/eupl
*
* Unless required by applicable law or agreed to in
writing, software distributed under the Licence is
distributed on an "AS IS" basis,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
express or implied.
* See the Licence for the specific language governing
permissions and limitations under the Licence.
'''

"""
Module for image enhancement to make identifiable / trackable features in an image clearer.
Much of this is based on the ideas and algorithms developed by:
Sam Van Stroud, Jamie McGowan, Augustin Marignier, Emily Nurse & Christian Gutschow
in a collaboration between UKAEA and University College London.
"""
import warnings

import cv2
import numpy as np
from scipy.optimize import curve_fit



def scale_to_8bit(image,cutoff=0.999):

    # If we already have an 8-bit image, don't do anything
    if image.dtype == np.uint8:
        return image

    # If we have a multi-channel image
    if len(image.shape) == 3:

        # Remove any transparency channel if there is one
        if image.shape[2] == 4:
            image = image[:, :, :-1]

        # If it's actually monochrome, throw away the colour
        if (np.diff(image.reshape(image.shape[2],-1),axis=0)==0).all():
            image = image[:,:,0]

    # Do the subtraction & normalisation as float to avoid
    # overflows or quantisation problems.
    image = image.astype(np.float32)
    image[np.isfinite(image) == False] = 0

    sorted = np.sort(image.flatten())
    maxval = sorted[int(sorted.size * cutoff)]
    minval = sorted[int(sorted.size * (1 - cutoff))]

    # Scale the image to its max & min
    image = np.maximum(0,image - minval)
    image = np.minimum(1,image / maxval)

    # Final 8 bit output
    return np.uint8(255 * image)


def enhance_image(image,target_msb=25,target_noise=500,tiles=(20,20),downsample=False,median=False,bilateral=False):
    """
    Enhance details in a given image. Used both for visual enhancement
    in the Calcam GUIs and as a pre-processing step for automatic camera
    movement detection.

    Parameters:
        image (np.ndarray)   : Image to enhance
        target_msb (float)   : Controls contrast enhancement. Higher numbers \
                               give more contrast enhancement.
        target_noise (float) : Controls level of de-noising. Lower values \
                               give more agressive de-noising.
        tiles (tuple of int) : Number of tiles in horizontal and vertical directions \
                               for local histogram equilisation.
        downsample (bool)    : Whether or not to downsample the image by a factor 2 \
                               using cv2.pyrDown(). Using this greatly increases the \
                               success of automatic point detection, but makes the images \
                               look worse by eye due to the lower resolution.

    Returns:

        np.ndarray : The processed image.
    """

    # Make sure we have an 8-bit unisnged int image.
    if image.dtype != np.uint8:
        image = scale_to_8bit(image)


    if len(image.shape) > 2:
        mono = False
        image_lab = cv2.cvtColor(image[:,:,:3],cv2.COLOR_RGB2LAB)
        image = image_lab[:,:,0]
    else:
        mono = True


    if downsample:
        image = cv2.pyrDown(image)
        if not mono:
            image_lab = cv2.resize(image_lab,(image.shape[1],image.shape[0]))
    else:
        target_noise = target_noise * 2

    if median:
        image = cv2.medianBlur(image,ksize=3)

    test_clip_lims = [1.,5.,10.]
    contrast = []

    for cliplim in test_clip_lims:
        contrast.append( local_contrast( cv2.createCLAHE(cliplim, tiles).apply(image), tiles) )

    result = image.copy()
    if max(contrast) - min(contrast) > 0:
        coefs = np.polyfit(contrast,test_clip_lims,2)
        best_cliplim = np.polyval(coefs,target_msb)
        if best_cliplim > 0:
            result = cv2.createCLAHE(best_cliplim, tiles).apply(image)


    if bilateral:
        result = cv2.bilateralFilter(result,d=-1,sigmaColor=25,sigmaSpace=25)

    starting_noise = cv2.Laplacian(image,cv2.CV_64F).var()
    nlm_win_size = int(np.mean([image.shape[0] / 100, image.shape[1] / 100]))

    if starting_noise > target_noise and nlm_win_size > 1 and min(image.shape) > 512:
        test_strengths = [1.,10.,20.,30.]

        noise = []
        for strength in test_strengths:
            noise.append(cv2.Laplacian(cv2.fastNlMeansDenoising(result,h=strength,searchWindowSize=nlm_win_size),cv2.CV_64F).var())

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            tanparams,_ = curve_fit(tan_shape,noise,test_strengths,p0=[1/np.max(noise),20,500,10],bounds=([1e-5,1e-4,-1000,-100],[3.14159/(np.max(noise) - np.min(noise)),100,1000,100]))

        best_strength = tan_shape(target_noise,*tanparams)

        if best_strength > 0:
            result = cv2.fastNlMeansDenoising(result,h=best_strength,searchWindowSize=nlm_win_size)
    if mono:
        result = np.tile(result[:,:,np.newaxis],(1,1,3))
    else:
        image_lab[:,:,0] = result
        result = cv2.cvtColor(image_lab,cv2.COLOR_LAB2RGB)

    return result



def local_contrast(image,tilegridsize=(20,20)):
    """
    Return a measure of the local contrast in a given image

    Parameters:
        image (np.ndarray)          : Image to process
        tilegridsize (tuple of int) : Number of tiles in horizontal and \
                                      vertical directions to split the image in to \
                                      for local contrast measurements.

    Returns:

        float: Local contrast parameter
    """
    tile_height = int(np.ceil(image.shape[0] / tilegridsize[1]))
    tile_width = int(np.ceil(image.shape[1] / tilegridsize[0]))
    sb = []

    for i in range(tilegridsize[1]):
        for j in range(tilegridsize[0]):
            sb.append(image[i*tile_height:(i+1)*tile_height,j*tile_width:(j+1)*tile_width].std())

    return np.nanmean(sb)


def tan_shape(x,xscale,yscale,xshift,yshift):
    """
    Tan fit function for use in image enhancement.

    Parameters:

        x (float or array) : Independent variable
        xscale (float)     : Horizontal scaling factor for tan function
        yscale (float)     : Vertical scaling factor for tan function
        xshift (float)     : Horizontal shift distance
        yshift (float)     : Vertical shift distance

    Returns:
        float : -yscale * (np.tan( (x - xshift)*xscale) + yshift)
    """
    return -yscale * (np.tan( (x - xshift)*xscale) + yshift)


'''
* Copyright 2015-2021 European Atomic Energy Community (EURATOM)
*
* Licensed under the EUPL, Version 1.1 or - as soon they
  will be approved by the European Commission - subsequent
  versions of the EUPL (the "Licence");
* You may not use this work except in compliance with the
  Licence.
* You may obtain a copy of the Licence at:
*
* https://joinup.ec.europa.eu/software/page/eupl
*
* Unless required by applicable law or agreed to in
  writing, software distributed under the Licence is
  distributed on an "AS IS" basis,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
  express or implied.
* See the Licence for the specific language governing
  permissions and limitations under the Licence.
'''


'''
A place to collect miscellaneous useful things which can
be used elsewhere in Calcam.
'''

import time
import getpass
import datetime
import socket
import importlib
import os
import sys
import subprocess

import numpy as np

# Current username and hostname
username = getpass.getuser()
hostname = socket.gethostname()


def get_formatted_time(timestamp=None):
    
    if timestamp is None:
        when = datetime.datetime.now()
    else:
        when = datetime.datetime.fromtimestamp(timestamp)
    
    return when.strftime('%H:%M on %Y-%m-%d')
    

def rotate_3d(vect,axis,angle):
    '''
    Rotate a given 3D coordinate about a given axis by a given angle.
	
    Parameters:
		
        vect (sequence) : 3 element sequence containing the point to rotate.
		
        axis (sequence) : 3 element sequence containing a vector defining the axis to rotate about.
		
        angle (float)   : Angle to rotate, in degrees
		
    Returns:
		
        np.array : 3 element array containing the rotated coordinates
	
    '''
    vect = np.array(vect,dtype=np.float64)
    vect_ = np.matrix(np.zeros([3,1]),dtype=np.float64)
    vect_[0,0] = vect[0]
    vect_[1,0] = vect[1]
    vect_[2,0] = vect[2]
    axis = np.array(axis)

    # Put angle in radians
    angle = angle * 3.14159 / 180.

    # Make sure the axis is normalised
    axis = axis / np.sqrt(np.sum(axis**2))

    # Make a rotation matrix!
    R = np.matrix(np.zeros([3,3]))
    R[0,0] = np.cos(angle) + axis[0]**2*(1 - np.cos(angle))
    R[0,1] = axis[0]*axis[1]*(1 - np.cos(angle)) - axis[2]*np.sin(angle)
    R[0,2] = axis[0]*axis[2]*(1 - np.cos(angle)) + axis[1]*np.sin(angle)
    R[1,0] = axis[1]*axis[0]*(1 - np.cos(angle)) + axis[2]*np.sin(angle)
    R[1,1] = np.cos(angle) + axis[1]**2*(1 - np.cos(angle))
    R[1,2] = axis[1]*axis[2]*(1 - np.cos(angle)) - axis[0]*np.sin(angle)
    R[2,0] = axis[2]*axis[0]*(1 - np.cos(angle)) - axis[1]*np.sin(angle)
    R[2,1] = axis[2]*axis[1]*(1 - np.cos(angle)) + axis[0]*np.sin(angle)
    R[2,2] = np.cos(angle) + axis[2]**2*(1 - np.cos(angle))

    return np.array( R * vect_)



class ColourCycle():
    '''
    A class to represent a colour cycle,
    used for when plotting / displaying multiple things.

    Note: this uses the same colours as the MatPlotLib 2 default
    colour cycle except with grey (C7) removed because usually grey doesn't
    stand out well enough.
    '''
    def __init__(self):

        self.colours = [(0.121,0.466,0.705),
                        (1,0.498,0.054),
                        (0.172,0.627,0.172),
                        (0.829,0.152,0.156),
                        (0.580,0.403,0.741),
                        (0.549,0.337,0.294),
                        (0.890,0.466,0.760),
                        (0.737,0.741,0.133),
                        (0.09,0.745,0.811),
                        ]

        self.extra_colours = []

        self.next_index = 0

        self.next = self.__next__

    def __iter__(self):
        return self

    def __next__(self):
        if len(self.extra_colours) > 0:
            return self.extra_colours.pop()
        else:
            col = self.colours[self.next_index]
            self.next_index = self.next_index + 1
            if self.next_index > len(self.colours) - 1:
                self.next_index = 0 
            return col

    def queue_colour(self,colour):
        self.extra_colours.insert(0,colour)




class DodgyDict():
    '''
    Custom dictionary-like storage class.
    Behaves more-or-less like a dictionary to the user but without the requirement
    that the keys are hashable. This is a bodge so I can do things like use
    QTreeWidgetItems as keys.
    '''
    def __init__(self):

        self.keylist = []
        self.itemlist = []
        self.iter_index = 0
        self.next = self.__next__

    def __getitem__(self,key):
        for i,ikey in enumerate(self.keylist):
            if key == ikey:
                return self.itemlist[i]
        raise KeyError('No item with key "{:}"'.format(key))

    def __setitem__(self,key,value):

        for i,ikey in enumerate(self.keylist):
            if key == ikey:
                self.itemlist[i] = value
                return

        self.keylist.append(key)
        self.itemlist.append(value)

    def __delitem__(self,key):
        for i,ikey in enumerate(self.keylist):
            if key == ikey:
                self.itemlist.remove(self.itemlist[i])
                self.keylist.remove(key)
                return

    def __iter__(self):
        self.iter_index = 0
        return self

    def __next__(self):
        if self.iter_index > len(self.keys()) - 1: 
            raise StopIteration
        else:
            self.iter_index += 1
            return (self.keylist[self.iter_index-1],self.itemlist[self.iter_index-1])

    def keys(self):
        return self.keylist
        
        
        
class LoopProgPrinter:
    '''
    A little object for telling the user how a long,
    loopy calculation is progressing, using stdout.
    
    Includes simple time to completion prediction.
    '''
    def __init__(self):
        
        self.starttime = None
        self.startdatetime = None
        self.frac_done = 0.
        
        # Config parameters
        self.wait_time = 5
        self.min_remaining_length = 5
        self.start_printed = False
        self.end_printed = False
        
        
    def update(self,status):
        
        if status is None:
            return
            
        try:
            float(status)
            self.frac_done = status
            if self.starttime is None:
                self.starttime = time.time()
                self.startdatetime = datetime.datetime.now()
        except (TypeError, ValueError):
            print(status)
            return
            

        elapsed_time = time.time() - self.starttime

        if elapsed_time > self.wait_time and not self.start_printed and self.frac_done > 0:

                est_time = (elapsed_time / self.frac_done)
                
                if est_time - elapsed_time > self.min_remaining_length:
                    est_time_string = ''
                    if est_time > 3600:
                        est_time_string = est_time_string + '{:.0f} hr '.format(np.floor(est_time/3600))
                    if est_time > 600:
                        est_time_string = est_time_string + '{:.0f} min.'.format((est_time - 3600*np.floor(est_time/3600))/60)
                    elif est_time > 59:
                        est_time_string = est_time_string + '{:.0f} min {:.0f} sec.'.format(np.floor(est_time/60),est_time % 60)
                    else:
                        est_time_string ='{:.0f} sec.'.format(est_time)
                    print(self.startdatetime.strftime('Started on:         %Y-%m-%d at %H:%M:%S'))
                    print('Estimated duration: {:s}'.format(est_time_string))
                    
                    self.start_printed = True
                    
        elif self.frac_done == 1. and not self.end_printed:
             
            tot_time = time.time() - self.starttime
            time_string = ''
            if tot_time > 3600:
                time_string = time_string + '{:.0f} hr '.format(np.floor(tot_time / 3600))
            if tot_time >= 59:
                time_string = time_string + '{:.0f} min '.format(np.floor( (tot_time - 3600*np.floor(tot_time / 3600))  / 60))
            time_string = time_string + '{:.0f} sec. '.format( tot_time - 60*np.floor(tot_time / 60) )
            print('Completed in:       {:s}'.format(time_string))
            
            
            self.end_printed = True


def bin_image(arr, factor,binfunc=np.mean):
    """
    Bin an image by the given factor.

    Parameters:

        arr     (array)    :  Input image / array for binning
        factor  (int)      :  Factor by which to bin, e.g. factor=2 would be 2x2 binning.
        binfunc (callable) :  Function to bin with, Numpy ufunc-style. Default is mean.

    Returns:

        Array with the binned image.

    """
    factor = int(factor)

    if arr.shape[0] % factor or arr.shape[1] % factor:
        raise ValueError('The binning factor {:d} is not an integer factor of the array dimensions {:d}x{:d}!'.format(factor,arr.shape[1],arr.shape[0]))

    shape = (arr.shape[0]//factor, factor,
             arr.shape[1]//factor, factor)

    if len(arr.shape) == 2:
        return binfunc(binfunc(arr.reshape(shape), axis=-1), axis=1)
    elif len(arr.shape) == 3:
        out = np.zeros((arr.shape[0] // factor, arr.shape[1] // factor) + arr.shape[2:], dtype=arr.dtype)
        for channel in range(arr.shape[2]):
            out[:,:,channel] = bin_image(arr[:,:,channel],factor,binfunc)
        return out


def import_source(source_path):
    """
    Import a python module from specified python source file,
    and return a reference to the imported module. Closely based
    on the importlib documentation.

    Parameters:
        source_path (str)   : Path to the source to be imported.

    Returns:

        python module

    """
    unload_source(source_path)

    if os.path.isdir(source_path) and os.path.isfile(os.path.join(source_path, '__init__.py')):
        path_elements = source_path.split(os.path.sep)
        try:
            path_elements.remove('')
        except ValueError:
            pass
        modname = path_elements[-1]
        source_path = os.path.join(source_path, '__init__.py')
    elif source_path.endswith('.py'):
        modname = os.path.split(source_path)[-1][:-3]
    else:
        raise FileNotFoundError('Specified path "{:s}" is not a python source file or package directory.'.format(source_path))

    # Check we're not going to create a name clash
    modlist = sys.modules.keys()
    modname_ = modname
    i = 0
    while modname_ in modlist:
        i += 1
        modname_ = modname + '_{:d}'.format(i)


    spec = importlib.util.spec_from_file_location(modname_, source_path)
    module = importlib.util.module_from_spec(spec)

    sys.modules[modname_] = module
    try:
        spec.loader.exec_module(module)
    except Exception:
        del sys.modules[modname_]
        raise

    return module


def unload_source(source_path):
    """
    Kill any references held by sys.modules for a module with a given source path.
    """
    # Derive what module name import_source would have given it
    if os.path.isdir(source_path) and os.path.isfile(os.path.join(source_path, '__init__.py')):
        path_elements = source_path.split(os.path.sep)
        try:
            path_elements.remove('')
        except ValueError:
            pass
        modname = path_elements[-1]
    elif source_path.endswith('.py'):
        modname = os.path.split(source_path)[-1][:-3]
    else:
        return

    # Find a list of modules it could be
    matching_modules = [mod for mod in sys.modules.keys() if modname in mod]

    # Only remove references to ones which really have the right source path
    for module in matching_modules:
        try:
            if sys.modules[module].__path__[0] == source_path:
                del sys.modules[module]
        except AttributeError:
            if source_path in sys.modules[module].__file__:
                del sys.modules[module]


def open_file(path):
    """
    Cross-platform opening of a path in the OS's default way.
    From https://stackoverflow.com/questions/17317219/is-there-an-platform-independent-equivalent-of-os-startfile
    """
    if sys.platform == "win32":
        os.startfile(path)
    else:
        opener = "open" if sys.platform == "darwin" else "xdg-open"
        subprocess.call([opener, path])


def get_contour_intersection(contour,linestart,lineend,debug=False):
    """
    Returns the intersection of a line from the x-point to given
    strike point, and the wall.

    Parameters:

        linestart (2 element array) : [R,Z] position of X-point
        lineend (2 element array) : [R,Z] position of "strike point"
        debug (bool) : Whether to plot positions for debugging purposes

    Returns:

        NaN if the line which goes through linestart and strike point does not intersect the wall

        S coordinate in mm of the wall intersection if it exists.

    """
    Rs = contour[:,0]
    Zs = contour[:,1]

    Re = np.roll(Rs, 1)
    Ze = np.roll(Zs, 1)

    hs = np.hstack((Rs[:, np.newaxis], Zs[:, np.newaxis], np.ones((Rs.size, 1))))
    he = np.hstack((Re[:, np.newaxis], Ze[:, np.newaxis], np.ones((Rs.size, 1))))
    l_s = np.cross(hs, he)

    l = np.tile( np.cross([linestart[0],linestart[1],1],[lineend[0],lineend[1],1]), (l_s.shape[0],1) )

    res = np.cross(l,l_s)
    with np.errstate(divide='ignore',invalid='ignore'):
        Ri = res[:,0] / res[:,2]
        Zi = res[:,1] / res[:,2]

    dir_s = np.hstack((Ri[:,np.newaxis]- linestart[0],Zi[:,np.newaxis]- linestart[1]))
    dir = np.tile((np.array(lineend) - np.array(linestart))[np.newaxis,:],(dir_s.shape[0],1))
    dp = np.sum(dir_s * dir, axis=1)

    valid = (dp >= 0) & (Ri >= np.minimum(Rs,Re)-1e-5) & (Ri <= np.maximum(Rs,Re)+1e-5) & (Zi >= np.minimum(Zs,Ze)-1e-5) & (Zi <= np.maximum(Zs,Ze)+1e-5)

    if np.count_nonzero(valid) > 0:

        if np.count_nonzero(valid) > 1:
            dr = Ri[valid] - linestart[0]
            dz = Zi[valid] - linestart[1]
            d = np.sqrt(dr**2 + dz**2)
            mindex = np.argmin(d)
        else:
            mindex = 0


        if debug:
            import matplotlib.pyplot as plt
            plt.plot(Rs,Zs,'k')
            plt.plot(linestart[0],linestart[1],'x',label='linestart',markersize=10)
            plt.plot(Ri[valid][mindex],Zi[valid][mindex],'o',label='Detected intersection')
            plt.plot(lineend[0],lineend[1],'s',label='lineend')
            plt.legend()
            plt.show()

        return Ri[valid][mindex],Zi[valid][mindex]
    else:

        if debug:
            import matplotlib.pyplot as plt
            plt.plot(Rs,Zs,'k')
            plt.plot(linestart[0],linestart[1],'x',markersize=10,label='linestart')
            plt.plot(Ri,Zi,'o',label='All intersections')
            plt.plot(lineend[0],lineend[1],'s',label='lineend')
            plt.xlim([2.2,3.1])
            plt.ylim([-1.8,-1.3])
            plt.legend()
            plt.show()

        return np.nan